# -*- coding: utf-8 -*-

"""Export the Bioregistry as a JSON-LD context."""

import json
from collections import ChainMap
from pathlib import Path
from textwrap import dedent
from typing import Mapping, Optional, Tuple

import click

import bioregistry
from bioregistry import get_pattern_map, get_prefix_map
from bioregistry.constants import (
    CONTEXT_BIOREGISTRY_PATH,
    EXPORT_CONTEXTS,
    SHACL_TURTLE_PATH,
)
from bioregistry.schema import Collection


@click.command()
def generate_contexts():
    """Generate various context files."""
    _context_prefix_maps()
    _collection_prefix_maps()

    prefix_map = get_prefix_map()
    pattern_map = get_pattern_map()
    _write_prefix_map(CONTEXT_BIOREGISTRY_PATH, prefix_map=prefix_map)
    _write_shacl(SHACL_TURTLE_PATH, prefix_map=prefix_map, pattern_map=pattern_map)


def _collection_prefix_maps():
    for collection in bioregistry.read_collections().values():
        name = collection.context
        if name is None:
            continue
        path_stub = EXPORT_CONTEXTS.joinpath(name)
        prefix_map = collection.as_prefix_map()
        pattern_map = get_pattern_map()
        _write_prefix_map(path_stub.with_suffix(".context.jsonld"), prefix_map=prefix_map)
        _write_shacl(
            path_stub.with_suffix(".context.ttl"), prefix_map=prefix_map, pattern_map=pattern_map
        )


def get_prescriptive_artifacts(
    key: str, include_synonyms: Optional[bool] = None
) -> Tuple[Mapping[str, str], Mapping[str, str]]:
    """Get a prescriptive prefix map."""
    context = bioregistry.get_context(key)
    if context is None:
        raise KeyError
    remapping = dict(
        ChainMap(
            *(
                bioregistry.get_registry_map(metaprefix)
                for metaprefix in context.prefix_priority or []
            ),
            context.prefix_remapping or {},
        )
    )
    include_synonyms = (
        include_synonyms if include_synonyms is not None else context.include_synonyms
    )
    prescriptive_prefix_map = get_prefix_map(
        remapping=remapping,
        priority=context.uri_prefix_priority,
        include_synonyms=include_synonyms,
        use_preferred=context.use_preferred,
    )
    prescriptive_pattern_map = get_pattern_map(
        remapping=remapping,
        include_synonyms=include_synonyms,
        use_preferred=context.use_preferred,
    )
    return prescriptive_prefix_map, prescriptive_pattern_map


def _context_prefix_maps():
    for key in bioregistry.read_contexts():
        prefix_map, pattern_map = get_prescriptive_artifacts(key)
        stub = EXPORT_CONTEXTS.joinpath(key)
        _write_prefix_map(stub.with_suffix(".context.jsonld"), prefix_map=prefix_map)
        _write_shacl(
            stub.with_suffix(".context.ttl"), prefix_map=prefix_map, pattern_map=pattern_map
        )

        if key == "obo":  # Special case, maybe put this in data model
            prefix_map, pattern_map = get_prescriptive_artifacts(key, include_synonyms=True)
            stub_double = EXPORT_CONTEXTS.joinpath(f"{key}_synonyms")
            _write_prefix_map(stub_double.with_suffix(".context.jsonld"), prefix_map=prefix_map)
            _write_shacl(
                stub_double.with_suffix(".context.ttl"),
                prefix_map=prefix_map,
                pattern_map=pattern_map,
            )


def _write_shacl(
    path: Path, *, prefix_map: Mapping[str, str], pattern_map: Optional[Mapping[str, str]] = None
) -> None:
    text = dedent(
        """\
        @prefix sh: <http://www.w3.org/ns/shacl#> .

        [
          sh:declare
        {entries}
        ] .
        """
    )
    entries = ",\n".join(
        f'    [ sh:prefix "{prefix}" ; sh:namespace "{uri_prefix}" ]'
        if not pattern_map or prefix not in pattern_map
        else f'    [ sh:prefix "{prefix}" ; sh:namespace "{uri_prefix}" ; sh:pattern "{pattern_map[prefix]}" ]'
        for prefix, uri_prefix in sorted(prefix_map.items())
    )
    path.write_text(text.format(entries=entries))


def _write_prefix_map(path: Path, *, prefix_map: Mapping[str, str]) -> None:
    with path.open("w") as file:
        json.dump(
            fp=file,
            indent=4,
            sort_keys=True,
            obj={
                "@context": prefix_map,
            },
        )


def collection_to_context_jsonlds(collection: Collection) -> str:
    """Get the JSON-LD context as a string from a given collection."""
    return json.dumps(collection.as_context_jsonld())


def get_obofoundry_prefix_map(include_synonyms: bool = False) -> Mapping[str, str]:
    """Get the OBO Foundry prefix map.

    :param include_synonyms: Should synonyms of each prefix also be included as additional prefixes, but with
        the same URL prefix?
    :return: A mapping from prefixes to prefix URLs.
    """
    return get_prescriptive_artifacts("obo")[0]


if __name__ == "__main__":
    generate_contexts()



# -*- coding: utf-8 -*-

"""Tests for downloading external data."""

import unittest

from bioregistry.external.obofoundry import get_obofoundry_example


class TestUtils(unittest.TestCase):
    """Test utilities."""

    def test_obolibrary_example(self):
        """Test looking up an example from the OBO Foundry PURL service configuration."""
        self.assertEqual("0011124", get_obofoundry_example("pcl"))



# -*- coding: utf-8 -*-

"""Download registry information from the OLS."""

import datetime
import enum
import json
import logging
from email.utils import parseaddr
from functools import lru_cache
from textwrap import dedent
from typing import Any, Mapping, Optional

from pydantic import BaseModel
from pystow.utils import download

from bioregistry.constants import DATA_DIRECTORY, EXTERNAL

__all__ = [
    "get_ols",
]

logger = logging.getLogger(__name__)

DIRECTORY = EXTERNAL / "ols"
DIRECTORY.mkdir(exist_ok=True, parents=True)
URL = "https://www.ebi.ac.uk/ols/api/ontologies?size=1000"
RAW_PATH = DIRECTORY / "raw.json"
PROCESSED_PATH = DIRECTORY / "processed.json"
OLS_PROCESSING = DATA_DIRECTORY / "processing_ols.json"

OLS_SKIP = {
    "co_321:root": "this is a mistake in the way OLS imports CO",
    "phi": "this is low quality and has no associated metadata",
    "epso": "can't figure out / not sure if still exists",
    "epio": "can't figure out / not sure if still exists",
}


def get_ols(force_download: bool = False):
    """Get the OLS registry."""
    if PROCESSED_PATH.exists() and not force_download:
        with PROCESSED_PATH.open() as file:
            return json.load(file)

    download(url=URL, path=RAW_PATH, force=True)
    with RAW_PATH.open() as file:
        data = json.load(file)

    if "next" in data["_links"]:
        raise NotImplementedError(
            "Need to implement paging since there are more entries than fit into one page"
        )

    processed = {}
    for ontology in data["_embedded"]["ontologies"]:
        ols_id = ontology["ontologyId"]
        # TODO better docs on how to maintain this file
        config = get_ols_processing().get(ols_id)
        if config is None:
            if ols_id not in OLS_SKIP:
                logger.warning("need to curate processing file for OLS prefix %s", ols_id)
            continue
        processed[ols_id] = _process(ontology, config)

    with PROCESSED_PATH.open("w") as file:
        json.dump(processed, file, indent=2, sort_keys=True)
    return processed


class VersionType(str, enum.Enum):
    """Types for OLS ontology versions."""

    date = "date"
    semver = "semver"
    other = "other"
    sequential = "sequential"
    garbage = "garbage"
    missing = "missing"


class OLSConfig(BaseModel):
    """Configuration for processing an OLS ontology."""

    prefix: str
    version_type: VersionType
    version_date_format: Optional[str]
    version_prefix: Optional[str]
    version_suffix: Optional[str]
    version_suffix_split: Optional[str]
    version_iri_prefix: Optional[str]
    version_iri_suffix: Optional[str]


def _process(  # noqa:C901
    ols_entry: Mapping[str, Any], processing: OLSConfig
) -> Optional[Mapping[str, str]]:
    ols_id = ols_entry["ontologyId"]
    config = ols_entry["config"]
    version_iri = config["versionIri"]

    title = config.get("title")
    if not title:
        title = config.get("localizedTitles", {}).get("en")

    description = config.get("description")
    if not description:
        description = config.get("localizedDescriptions", {}).get("en")

    rv = {
        "prefix": ols_id,
        "name": title,
        "download": _clean_url(config["fileLocation"]),
        "version.iri": _clean_url(version_iri),
        "description": description,
        "homepage": _clean_url(config["homepage"]),
    }

    email = config.get("mailingList")
    if email:
        name, email = parseaddr(email)
        if email.startswith("//"):
            logger.debug("[%s] invalid email address: %s", ols_id, config["mailingList"])
        else:
            rv["contact"] = email

    license_value = config.get("annotations", {}).get("license", [None])[0]
    if license_value in {"Unspecified", "Unspecified"}:
        license_value = None
    if not license_value:
        logger.info("[%s] missing license in OLS. Contact: %s", ols_id, config["mailingList"])
    rv["license"] = license_value

    version = config.get("version")
    if version is None and processing.version_iri_prefix:
        if not version_iri.startswith(processing.version_iri_prefix):
            logger.info("[%s] version IRI does not start with appropriate prefix", ols_id)
        else:
            version = version_iri[len(processing.version_iri_prefix) :]
            if processing.version_iri_suffix:
                version = version[: -len(processing.version_iri_suffix)]

    if version is None:
        logger.info(
            "[%s] missing version in OLS. Contact: %s, consider version.iri %s",
            ols_id,
            config["mailingList"],
            version_iri,
        )
    else:
        if version != version.strip():
            logger.info(
                "[%s] extra whitespace in version: %s. Contact: %s",
                ols_id,
                version,
                config["mailingList"],
            )
            version = version.strip()

        version_prefix = processing.version_prefix
        if version_prefix:
            if not version.startswith(version_prefix):
                raise ValueError(
                    dedent(
                        f"""\
                [{ols_id}] version "{version}" does not start with prefix "{version_prefix}".
                Update the ["{ols_id}"]["prefix"] entry in the OLS processing configuration.
                """
                    )
                )
            version = version[len(version_prefix) :]
        if processing.version_suffix_split:
            version = version.split()[0]
        version_suffix = processing.version_suffix
        if version_suffix:
            if not version.endswith(version_suffix):
                raise ValueError(
                    f"[{ols_id}] version {version} does not end with prefix {version_suffix}"
                )
            version = version[: -len(version_suffix)]

        version_type = processing.version_type
        version_date_fmt = processing.version_date_format
        if version_date_fmt:
            if version_date_fmt in {"%Y-%d-%m"}:
                logger.info(
                    "[%s] confusing date format: %s. Contact: %s",
                    ols_id,
                    version_date_fmt,
                    config["mailingList"],
                )
            try:
                version = datetime.datetime.strptime(version, version_date_fmt).strftime("%Y-%m-%d")
            except ValueError:
                logger.info("[%s] wrong format for version %s", ols_id, version)
        elif not version_type:
            logger.info("[%s] no type for version %s", ols_id, version)

    rv["version"] = version

    rv = {k: v for k, v in rv.items() if v}
    return rv


def _clean_url(url: Optional[str]) -> Optional[str]:
    if url is None:
        return url
    if "CO_" in url and url.startswith("http://127.0.0.1:5900"):
        return "https://cropontology.org" + url[len("http://127.0.0.1:5900") :]
    return url


@lru_cache(maxsize=1)
def get_ols_processing() -> Mapping[str, OLSConfig]:
    """Get OLS processing configurations."""
    with OLS_PROCESSING.open() as file:
        data = json.load(file)
    return {record["prefix"]: OLSConfig(**record) for record in data["configurations"]}


if __name__ == "__main__":
    print(len(get_ols(force_download=True)))  # noqa:T201



"""Export the schema diagram."""

import click
from networkx.drawing.nx_agraph import to_agraph

from bioregistry.constants import SCHEMA_PDF_PATH, SCHEMA_SVG_PATH
from bioregistry.schema.constants import get_schema_nx


@click.command()
def schema_export():
    """Export the schema diagram."""
    agraph = to_agraph(get_schema_nx())
    agraph.layout(prog="dot")
    agraph.draw(SCHEMA_SVG_PATH)
    agraph.draw(SCHEMA_PDF_PATH)


if __name__ == "__main__":
    schema_export()



# -*- coding: utf-8 -*-

"""Workflows for aligning external registries."""


from .biocontext import BioContextAligner
from .biolink import BiolinkAligner
from .bioportal import BioPortalAligner
from .cellosaurus import CellosaurusAligner
from .cheminf import ChemInfAligner
from .fairsharing import FairsharingAligner
from .go import GoAligner
from .miriam import MiriamAligner
from .n2t import N2TAligner
from .ncbi import NcbiAligner
from .obofoundry import OBOFoundryAligner
from .ols import OLSAligner
from .ontobee import OntobeeAligner
from .prefixcommons import PrefixCommonsAligner
from .uniprot import UniProtAligner
from .utils import Aligner
from .wikidata import WikidataAligner

__all__ = [
    # Abstract
    "Aligner",
    "ALIGNERS",
    # Concrete
    "BioContextAligner",
    "BiolinkAligner",
    "BioPortalAligner",
    "CellosaurusAligner",
    "ChemInfAligner",
    "FairsharingAligner",
    "GoAligner",
    "MiriamAligner",
    "N2TAligner",
    "NcbiAligner",
    "OBOFoundryAligner",
    "OLSAligner",
    "OntobeeAligner",
    "PrefixCommonsAligner",
    "UniProtAligner",
    "WikidataAligner",
]

ALIGNERS = [sc for sc in Aligner.__subclasses__() if hasattr(sc, "key")]



"""This script suggests prefixes that might be providers for uniprot."""

import click
from tabulate import tabulate

import bioregistry

FALSE_POSITIVES = {
    "panther.pathway",
    "panther.pthcmp",
    "protclustdb",
}


@click.command()
def _main():
    rows = []
    for prefix, resource in bioregistry.read_registry().items():
        if prefix == "uniprot" or prefix in FALSE_POSITIVES:
            continue
        example = resource.get_example()
        if example is None:
            continue
        if resource.provides:
            continue
        if bioregistry.is_known_identifier("uniprot", example):
            rows.append((prefix, example, bioregistry.get_uri_format(prefix)))
    click.echo(tabulate(rows, headers=["prefix", "example", "uri_format"]))


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Linting functions."""

import click

from bioregistry.schema_utils import (
    read_collections,
    read_contexts,
    read_metaregistry,
    read_registry,
    write_collections,
    write_contexts,
    write_metaregistry,
    write_registry,
)


@click.command()
def lint():
    """Run the lint commands."""
    registry = read_registry()
    for resource in registry.values():
        if resource.synonyms:
            resource.synonyms = sorted(resource.synonyms)
    write_registry(registry)
    write_collections(read_collections())
    write_metaregistry(read_metaregistry())
    write_contexts(read_contexts())


if __name__ == "__main__":
    lint()



# -*- coding: utf-8 -*-

"""Align the BioContext with the Bioregistry."""

from typing import Any, Dict, Mapping, Sequence

from bioregistry.align.utils import Aligner
from bioregistry.constants import URI_FORMAT_KEY
from bioregistry.external.biocontext import get_biocontext

__all__ = [
    "BioContextAligner",
]


class BioContextAligner(Aligner):
    """Aligner for BioContext."""

    key = "biocontext"
    getter = get_biocontext
    curation_header = [URI_FORMAT_KEY, "identifiers", "purl"]

    def get_skip(self) -> Mapping[str, str]:
        """Get entries for BioContext that should be skipped."""
        return {
            "fbql": "not a real resource, as far as I can tell",
        }

    def prepare_external(self, external_id, external_entry) -> Dict[str, Any]:
        """Prepare BioContext data to be added to the BioContext for each BioPortal registry entry."""
        uri_format = external_entry[URI_FORMAT_KEY].strip()
        return {
            URI_FORMAT_KEY: uri_format,
            "is_identifiers": uri_format.startswith("http://identifiers.org"),
            "is_obo": uri_format.startswith("http://purl.obolibrary.org"),
        }

    def get_curation_row(self, external_id, external_entry) -> Sequence[str]:
        """Prepare curation rows for unaligned BioContext registry entries."""
        formatter = external_entry[URI_FORMAT_KEY]
        return [
            formatter,
            formatter.startswith("http://identifiers.org"),
            formatter.startswith("http://purl.obolibrary.org"),
        ]


if __name__ == "__main__":
    BioContextAligner.align()



# -*- coding: utf-8 -*-

"""Tests for checking the integrity of the contexts."""

import json
import unittest

import bioregistry
from bioregistry.constants import CONTEXTS_PATH
from bioregistry.utils import extended_encoder


class TestContexts(unittest.TestCase):
    """A test case for checking the integrity of the contexts."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.contexts = bioregistry.read_contexts()
        self.collection_keys = {
            collection.context: key
            for key, collection in bioregistry.read_collections().items()
            if collection.context
        }
        self.valid_metaprefixes = set(bioregistry.read_metaregistry()) | {"default"}
        self.valid_prefixes = set(bioregistry.read_registry())

    def test_linted(self):
        """Test the context file is linted."""
        text = CONTEXTS_PATH.read_text(encoding="utf-8")
        linted_text = json.dumps(
            json.loads(text),
            indent=2,
            sort_keys=True,
            ensure_ascii=False,
            default=extended_encoder,
        )
        self.assertEqual(linted_text, text)

    def test_data(self):
        """Test the data integrity."""
        for key, context in self.contexts.items():
            self.assertNotIn(
                key,
                set(self.collection_keys),
                msg=f"Context has same key as context assigned by collection {self.collection_keys.get(key)}",
            )

            for maintainer in context.maintainers:
                self.assertIsNotNone(maintainer.name)
                # self.assertIsNotNone(maintainer.email, msg=f"{maintainer.name} is missing an email")
                self.assertIsNotNone(
                    maintainer.github, msg=f"{maintainer.name} is missing a GitHub handle"
                )
                self.assertIsNotNone(maintainer.orcid, msg=f"{maintainer.name} is missing an ORCID")
                self.assertRegex(maintainer.orcid, "^\\d{4}-\\d{4}-\\d{4}-\\d{3}(\\d|X)$")

            for metaprefix in context.uri_prefix_priority or []:
                self.assertIn(metaprefix, self.valid_metaprefixes)
            for metaprefix in context.prefix_priority or []:
                self.assertIn(metaprefix, self.valid_metaprefixes)
            remapping = context.prefix_remapping or {}
            _valid_remapping_prefixes = set(
                bioregistry.get_prefix_map(
                    priority=context.uri_prefix_priority,
                    use_preferred=context.use_preferred,
                )
            )
            for prefix in remapping:
                # Currently this allows overwriting of existing prefixes
                self.assertIn(prefix, _valid_remapping_prefixes)

            _valid_custom_prefixes = set(
                bioregistry.get_prefix_map(
                    remapping=remapping,
                    priority=context.uri_prefix_priority,
                    use_preferred=context.use_preferred,
                )
            )
            invalid_custom_prefixes = {
                prefix
                for prefix in (context.custom_prefix_map or {})
                if prefix not in _valid_custom_prefixes
            }
            self.assertEqual(
                0,
                len(invalid_custom_prefixes),
                msg=f"""

    All prefixes in the custom prefix mapping should either be canonical prefixes or generated by the prefix remapping

    Invalid prefixes: {", ".join(sorted(invalid_custom_prefixes))}
""",
            )



# -*- coding: utf-8 -*-

"""Assert regular expression patterns for common identifier styles."""

import re

import click
import tabulate

import bioregistry


def _examples(resource):
    rv = []
    example = resource.get_example()
    if example:
        rv.append(example)
    rv.extend(resource.example_extras or [])
    return rv


@click.command()
def _main():
    registry = bioregistry.read_registry()
    rows = []
    for prefix, resource in registry.items():
        pattern = resource.get_pattern()
        if pattern:
            continue
        examples = _examples(resource)
        if not examples:
            continue
        example = examples[0]

        if len(example) > 10:
            continue  # skip long examples
        if example.startswith("0"):
            pattern = f"^\\d{{{len(example)}}}$"
            if not all(re.match(pattern, e) for e in examples):
                click.echo(f"Check prefix: {prefix}")
                continue
            resource.pattern = pattern
            rows.append((prefix, example, resource.pattern))
        elif example.isnumeric():
            pattern = "^\\d+$"
            if not all(re.match(pattern, e) for e in examples):
                click.echo(f"Check prefix: {prefix}")
                continue
            resource.pattern = pattern
            rows.append((prefix, example, resource.pattern))
        else:
            rows.append((prefix, example, "???"))

    click.echo(tabulate.tabulate(rows, headers=["prefix", "example", "pattern"]))
    bioregistry.write_registry(registry)


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""A class-based client to a metaregistry."""

import logging
from collections import defaultdict
from functools import lru_cache
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from .constants import (
    BIOREGISTRY_REMOTE_URL,
    IDENTIFIERS_ORG_URL_PREFIX,
    LINK_PRIORITY,
    MIRIAM_BLACKLIST,
)
from .license_standardizer import standardize_license
from .schema import Registry, Resource, sanitize_model
from .schema_utils import (
    _registry_from_path,
    read_metaregistry,
    read_registry,
    write_registry,
)
from .utils import NormDict, _norm, curie_to_str

__all__ = [
    "Manager",
    "manager",
]

logger = logging.getLogger(__name__)


def _synonym_to_canonical(registry: Mapping[str, Resource]) -> NormDict:
    """Return a mapping from several variants of each synonym to the canonical namespace."""
    norm_synonym_to_key = NormDict()

    for identifier, resource in registry.items():
        norm_synonym_to_key[identifier] = identifier
        for synonym in resource.synonyms or []:
            norm_synonym_to_key[synonym] = identifier

        for metaprefix in ("miriam", "ols", "obofoundry", "go"):
            external = resource.get_external(metaprefix)
            if external is None:
                continue
            external_prefix = external.get("prefix")
            if external_prefix is None:
                continue
            if external_prefix not in norm_synonym_to_key:
                logger.debug(f"[{identifier}] missing potential synonym: {external_prefix}")

    return norm_synonym_to_key


class Manager:
    """A manager for functionality related to a metaregistry."""

    registry: Dict[str, Resource]
    metaregistry: Dict[str, Registry]

    def __init__(
        self,
        registry: Optional[Mapping[str, Resource]] = None,
        metaregistry: Optional[Mapping[str, Registry]] = None,
    ):
        """Instantiate a registry manager.

        :param registry: A custom registry. If none given, defaults to the Bioregistry.
        :param metaregistry: A custom metaregistry. If none, defaults to the Bioregistry's metaregistry.
        """
        self.registry = dict(read_registry() if registry is None else registry)
        self.synonyms = _synonym_to_canonical(self.registry)

        self.metaregistry = dict(read_metaregistry() if metaregistry is None else metaregistry)

        canonical_for = defaultdict(list)
        provided_by = defaultdict(list)
        has_parts = defaultdict(list)
        for prefix, resource in self.registry.items():
            if resource.has_canonical:
                canonical_for[resource.has_canonical].append(prefix)
            if resource.provides:
                provided_by[resource.provides].append(prefix)
            if resource.part_of:
                has_parts[resource.part_of].append(prefix)
        self.canonical_for = dict(canonical_for)
        self.provided_by = dict(provided_by)
        self.has_parts = dict(has_parts)

    @classmethod
    def from_path(cls, path: Union[str, Path]) -> "Manager":
        """Load a manager from the given path."""
        return cls(_registry_from_path(path))

    def write_registry(self):
        """Write the registry."""
        write_registry(self.registry)

    def normalize_prefix(self, prefix: str) -> Optional[str]:
        """Get the normalized prefix, or return None if not registered.

        :param prefix: The prefix to normalize, which could come from Bioregistry,
            OBO Foundry, OLS, or any of the curated synonyms in the Bioregistry
        :returns: The canonical Bioregistry prefix, it could be looked up. This
            will usually take precedence: MIRIAM, OBO Foundry / OLS, Custom except
            in a few cases, such as NCBITaxon.
        """
        return self.synonyms.get(prefix)

    def get_resource(self, prefix: str) -> Optional[Resource]:
        """Get the Bioregistry entry for the given prefix.

        :param prefix: The prefix to look up, which is normalized with :func:`normalize_prefix`
            before lookup in the Bioregistry
        :returns: The Bioregistry entry dictionary, which includes several keys cross-referencing
            other registries when available.
        """
        norm_prefix = self.normalize_prefix(prefix)
        if norm_prefix is None:
            return None
        return self.registry.get(norm_prefix)

    def parse_curie(self, curie: str, sep: str = ":") -> Union[Tuple[str, str], Tuple[None, None]]:
        """Parse a CURIE and normalize its prefix and identifier."""
        try:
            prefix, identifier = curie.split(sep, 1)
        except ValueError:
            return None, None
        return self.normalize_parsed_curie(prefix, identifier)

    def normalize_curie(self, curie: str, sep: str = ":") -> Optional[str]:
        """Normalize the prefix and identifier in the CURIE."""
        prefix, identifier = self.parse_curie(curie, sep=sep)
        if prefix is None or identifier is None:
            return None
        return curie_to_str(prefix, identifier)

    def normalize_parsed_curie(
        self, prefix: str, identifier: str
    ) -> Union[Tuple[str, str], Tuple[None, None]]:
        """Normalize a prefix/identifier pair.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A normalized prefix/identifier pair, conforming to Bioregistry standards. This means no redundant
            prefixes or bananas, all lowercase.
        """
        norm_prefix = self.normalize_prefix(prefix)
        if not norm_prefix:
            return None, None
        resource = self.registry[norm_prefix]
        norm_identifier = resource.standardize_identifier(identifier, prefix=prefix)
        return norm_prefix, norm_identifier

    @lru_cache(maxsize=None)  # noqa:B019
    def get_registry_map(self, metaprefix: str) -> Dict[str, str]:
        """Get a mapping from the Bioregistry prefixes to prefixes in another registry."""
        return dict(self._iter_registry_map(metaprefix))

    @lru_cache(maxsize=None)  # noqa:B019
    def get_registry_invmap(self, metaprefix: str, normalize: bool = False) -> Dict[str, str]:
        """Get a mapping from prefixes in another registry to Bioregistry prefixes."""
        if normalize:
            return {
                _norm(external_prefix): prefix
                for prefix, external_prefix in self._iter_registry_map(metaprefix)
            }
        return {
            external_prefix: prefix
            for prefix, external_prefix in self._iter_registry_map(metaprefix)
        }

    def _iter_registry_map(self, metaprefix: str) -> Iterable[Tuple[str, str]]:
        for prefix, resource in self.registry.items():
            mapped_prefix = resource.get_mapped_prefix(metaprefix)
            if mapped_prefix is not None:
                yield prefix, mapped_prefix

    def get_mapped_prefix(self, prefix: str, metaprefix: str) -> Optional[str]:
        """Get the prefix mapped into another registry."""
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        return resource.get_mapped_prefix(metaprefix)

    def get_external(self, prefix: str, metaprefix: str) -> Mapping[str, Any]:
        """Get the external data for the entry."""
        entry = self.get_resource(prefix)
        if entry is None:
            return {}
        return entry.get_external(metaprefix)

    def get_versions(self) -> Mapping[str, str]:
        """Get a map of prefixes to versions."""
        rv = {}
        for prefix, resource in self.registry.items():
            version = resource.get_version()
            if version is not None:
                rv[prefix] = version
        return rv

    def get_uri_format(self, prefix, priority: Optional[Sequence[str]] = None) -> Optional[str]:
        """Get the URI format string for the given prefix, if it's available."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_uri_format(priority=priority)

    def get_uri_prefix(self, prefix, priority: Optional[Sequence[str]] = None) -> Optional[str]:
        """Get a well-formed URI prefix, if available."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_uri_prefix(priority=priority)

    def get_name(self, prefix: str) -> Optional[str]:
        """Get the name for the given prefix, it it's available."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_name()

    def get_preferred_prefix(self, prefix: str) -> Optional[str]:
        """Get the preferred prefix (e.g., with stylization) if it exists."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_preferred_prefix()

    def get_pattern(self, prefix: str) -> Optional[str]:
        """Get the pattern for the given prefix, if it's available."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_pattern()

    def get_synonyms(self, prefix: str) -> Optional[Set[str]]:
        """Get the synonyms for a given prefix, if available."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_synonyms()

    def get_example(self, prefix: str) -> Optional[str]:
        """Get an example identifier, if it's available."""
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_example()

    def has_no_terms(self, prefix: str) -> bool:
        """Get if the entry has been annotated to not have own terms."""
        entry = self.get_resource(prefix)
        if entry is None or entry.no_own_terms is None:
            return False
        return entry.no_own_terms

    def is_deprecated(self, prefix: str) -> bool:
        """Return if the given prefix corresponds to a deprecated resource."""
        entry = self.get_resource(prefix)
        if entry is None:
            return False
        return entry.is_deprecated()

    def get_pattern_map(
        self,
        *,
        include_synonyms: bool = False,
        remapping: Optional[Mapping[str, str]] = None,
        use_preferred: bool = False,
    ) -> Mapping[str, str]:
        """Get a mapping from prefixes to their regular expression patterns.

        :param include_synonyms: Should synonyms of each prefix also be included as additional prefixes, but with
            the same URI prefix?
        :param remapping: A mapping from prefixes to preferred prefixes.
        :param use_preferred: Should preferred prefixes be used? Set this to true if you're in the OBO context.
        :return: A mapping from prefixes to regular expression pattern strings.
        """
        it = self._iter_pattern_map(include_synonyms=include_synonyms, use_preferred=use_preferred)
        if not remapping:
            return dict(it)
        return {remapping.get(prefix, prefix): uri_prefix for prefix, uri_prefix in it}

    def _iter_pattern_map(
        self,
        *,
        include_synonyms: bool = False,
        use_preferred: bool = False,
    ) -> Iterable[Tuple[str, str]]:
        for prefix, resource in self.registry.items():
            pattern = resource.get_pattern()
            if pattern is None:
                continue
            if use_preferred:
                preferred_prefix = resource.get_preferred_prefix()
                if preferred_prefix is not None:
                    prefix = preferred_prefix
            yield prefix, pattern
            if include_synonyms:
                for synonym in resource.get_synonyms():
                    yield synonym, pattern

    def get_prefix_map(
        self,
        *,
        priority: Optional[Sequence[str]] = None,
        include_synonyms: bool = False,
        remapping: Optional[Mapping[str, str]] = None,
        use_preferred: bool = False,
    ) -> Mapping[str, str]:
        """Get a mapping from Bioregistry prefixes to their URI prefixes .

        :param priority: A priority list for how to generate URI prefixes.
        :param include_synonyms: Should synonyms of each prefix also be included as additional prefixes, but with
            the same URI prefix?
        :param remapping: A mapping from Bioregistry prefixes to preferred prefixes.
        :param use_preferred: Should preferred prefixes be used? Set this to true if you're in the OBO context.
        :return: A mapping from prefixes to URI prefixes.
        """
        it = self._iter_prefix_map(
            priority=priority, include_synonyms=include_synonyms, use_preferred=use_preferred
        )
        if not remapping:
            return dict(it)
        return {remapping.get(prefix, prefix): uri_prefix for prefix, uri_prefix in it}

    def _iter_prefix_map(
        self,
        *,
        priority: Optional[Sequence[str]] = None,
        include_synonyms: bool = False,
        use_preferred: bool = False,
    ) -> Iterable[Tuple[str, str]]:
        for prefix, resource in self.registry.items():
            uri_prefix = resource.get_uri_prefix(priority=priority)
            if uri_prefix is None:
                continue
            if use_preferred:
                preferred_prefix = resource.get_preferred_prefix()
                if preferred_prefix is not None:
                    prefix = preferred_prefix
            yield prefix, uri_prefix
            if include_synonyms:
                for synonym in resource.get_synonyms():
                    yield synonym, uri_prefix

    def get_prefix_list(self, **kwargs) -> List[Tuple[str, str]]:
        """Get the default priority prefix list."""
        #: A prefix map in reverse sorted order based on length of the URI prefix
        #: in order to avoid conflicts of sub-URIs (thanks to Nico Matentzoglu for the idea)
        return prepare_prefix_list(self.get_prefix_map(**kwargs))

    def get_curie_pattern(self, prefix: str) -> Optional[str]:
        """Get the CURIE pattern for this resource.

        :param prefix: The prefix to look up
        :return: The regular expression pattern to match CURIEs against
        """
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        pattern = resource.get_pattern()
        if pattern is None:
            return None
        p = resource.get_preferred_prefix() or prefix
        p = p.replace(".", "\\.")
        return f"^{p}:{pattern.lstrip('^')}"

    def rasterize(self):
        """Build a dictionary representing the fully constituted registry."""
        return {
            prefix: sanitize_model(resource)
            for prefix, resource in self._rasterized_registry().items()
        }

    def _rasterized_registry(self) -> Mapping[str, Resource]:
        return {
            prefix: self._rasterized_resource(prefix, resource)
            for prefix, resource in self.registry.items()
        }

    def _rasterized_resource(self, prefix: str, resource: Resource) -> Resource:
        return Resource(
            prefix=resource.prefix,
            preferred_prefix=resource.get_preferred_prefix() or prefix,
            name=resource.get_name(),
            description=resource.get_description(),
            pattern=resource.get_pattern(),
            homepage=resource.get_homepage(),
            license=resource.get_license(),
            version=resource.get_version(),
            synonyms=resource.get_synonyms(),
            repository=resource.get_repository(),
            # Downloads
            download_obo=resource.get_download_obo(),
            download_json=resource.get_download_obograph(),
            download_owl=resource.get_download_owl(),
            # Registry properties
            example=resource.get_example(),
            example_extras=resource.example_extras,
            uri_format=resource.get_uri_format(),
            providers=resource.get_extra_providers(),
            # Comments
            comment=resource.comment,
            references=resource.references,
            # MIRIAM compatibility
            banana=resource.get_banana(),
            namespace_in_lui=resource.get_namespace_in_lui(),
            # Provenance
            contact=resource.get_contact(),
            contributor=resource.contributor,
            reviewer=resource.reviewer,
            # Ontology Relations
            part_of=resource.part_of,
            provides=resource.provides,
            has_canonical=resource.has_canonical,
            appears_in=self.get_appears_in(prefix),
            depends_on=self.get_depends_on(prefix),
            mappings=resource.get_mappings(),
            # Ontology Properties
            deprecated=resource.is_deprecated(),
            no_own_terms=resource.no_own_terms,
            proprietary=resource.proprietary,
        )

    def get_license_conflicts(self):
        """Get license conflicts."""
        conflicts = []
        for prefix, entry in self.registry.items():
            override = entry.license
            obo_license = entry.get_external("obofoundry").get("license")
            ols_license = entry.get_external("ols").get("license")
            if 2 > sum(license_ is not None for license_ in (override, obo_license, ols_license)):
                continue  # can't be a conflict if all none or only 1 is available
            obo_norm = standardize_license(obo_license)
            ols_norm = standardize_license(ols_license)
            first, *rest = [
                norm_license
                for norm_license in (override, obo_norm, ols_norm)
                if norm_license is not None
            ]
            if any(first != element for element in rest):
                conflicts.append((prefix, override, obo_license, ols_license))
        return conflicts

    def get_appears_in(self, prefix: str) -> Optional[List[str]]:
        """Return a list of resources that this resources (has been annotated to) depends on.

        This is complementary to :func:`get_depends_on`.

        :param prefix: The prefix to look up
        :returns: The list of resources this prefix has been annotated to appear in. This
            list could be incomplete, since curation of these fields can easily get out
            of sync with curation of the resource itself. However, false positives should
            be pretty rare.
        """
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        rv = list(resource.appears_in or [])
        rv.extend(self._get_obo_list(prefix=prefix, resource=resource, key="appears_in"))
        return sorted(set(rv))

    def get_depends_on(self, prefix: str) -> Optional[List[str]]:
        """Return a list of resources that this resources (has been annotated to) depends on.

        This is complementary to :func:`get_appears_in`.

        :param prefix: The prefix to look up
        :returns: The list of resources this prefix has been annotated to depend on. This
            list could be incomplete, since curation of these fields can easily get out
            of sync with curation of the resource itself. However, false positives should
            be pretty rare.

        >>> from bioregistry import manager
        >>> assert "bfo" in manager.get_depends_on("foodon")
        """
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        rv = list(resource.depends_on or [])
        rv.extend(self._get_obo_list(prefix=prefix, resource=resource, key="depends_on"))
        return sorted(set(rv))

    def _get_obo_list(self, *, prefix: str, resource: Resource, key: str) -> List[str]:
        rv = []
        for obo_prefix in resource.get_external("obofoundry").get(key, []):
            canonical_prefix = self.lookup_from("obofoundry", obo_prefix, normalize=True)
            if canonical_prefix is None:
                logger.warning("[%s] could not map OBO %s: %s", prefix, key, obo_prefix)
            else:
                rv.append(canonical_prefix)
        return rv

    def lookup_from(
        self, metaprefix: str, metaidentifier: str, normalize: bool = False
    ) -> Optional[str]:
        """Get the bioregistry prefix from an external prefix.

        :param metaprefix: The key for the external registry
        :param metaidentifier: The prefix in the external registry
        :param normalize: Should external prefixes be normalized during lookup (e.g., lowercased)
        :return: The bioregistry prefix (if it can be mapped)

        >>> from bioregistry import manager
        >>> manager.lookup_from("obofoundry", "GO")
        'go'
        >>> manager.lookup_from("obofoundry", "go")
        None
        >>> manager.lookup_from("obofoundry", "go", normalize=True)
        'go'
        """
        external_id_to_bioregistry_id = self.get_registry_invmap(metaprefix, normalize=normalize)
        return external_id_to_bioregistry_id.get(
            _norm(metaidentifier) if normalize else metaidentifier
        )

    def get_has_canonical(self, prefix: str) -> Optional[str]:
        """Get the canonical prefix."""
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        return resource.has_canonical

    def get_canonical_for(self, prefix: str) -> Optional[List[str]]:
        """Get the prefixes for which this is annotated as canonical."""
        norm_prefix = self.normalize_prefix(prefix)
        if norm_prefix is None:
            return None
        return self.canonical_for.get(norm_prefix, [])

    def get_provides_for(self, prefix: str) -> Optional[str]:
        """Get the resource that the given prefix provides for, or return none if not a provider."""
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        return resource.provides

    def get_provided_by(self, prefix: str) -> Optional[List[str]]:
        """Get the resources that provide for the given prefix, or return none if the prefix can't be looked up."""
        norm_prefix = self.normalize_prefix(prefix)
        if norm_prefix is None:
            return None
        return self.provided_by.get(norm_prefix, [])

    def get_part_of(self, prefix: str) -> Optional[str]:
        """Get the parent resource, if annotated."""
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        return resource.part_of

    def get_has_parts(self, prefix: str) -> Optional[List[str]]:
        """Get the children resources, if annotated."""
        norm_prefix = self.normalize_prefix(prefix)
        if norm_prefix is None:
            return None
        return self.has_parts.get(norm_prefix, [])

    def get_bioregistry_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get a Bioregistry link.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A link to the Bioregistry resolver
        """
        norm_prefix, norm_identifier = self.normalize_parsed_curie(prefix, identifier)
        if norm_prefix is None:
            return None
        return f"{BIOREGISTRY_REMOTE_URL.rstrip()}/{norm_prefix}:{norm_identifier}"

    def get_default_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the default URL for the given CURIE.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A IRI string corresponding to the default provider, if available.

        >>> from bioregistry import manager
        >>> manager.get_default_iri('chebi', '24867')
        'https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:24867'
        """
        entry = self.get_resource(prefix)
        if entry is None:
            return None
        return entry.get_default_uri(identifier)

    def get_miriam_curie(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the identifiers.org CURIE for the given CURIE."""
        resource = self.get_resource(prefix)
        if resource is None:
            return None
        miriam_prefix = self.get_mapped_prefix(prefix, "miriam")
        if miriam_prefix is None or miriam_prefix in MIRIAM_BLACKLIST:
            return None
        banana = resource.get_banana()
        if banana:
            if identifier.startswith(f"{banana}:"):
                return identifier
            else:
                return f"{banana}:{identifier}"
        elif resource.get_namespace_in_lui():
            if identifier.startswith(prefix.upper()):
                return identifier
            else:
                return f"{prefix.upper()}:{identifier}"
        else:
            return f"{miriam_prefix}:{identifier}"

    def get_miriam_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the identifiers.org URL for the given CURIE.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A IRI string corresponding to the Identifiers.org, if the prefix exists and is
            mapped to MIRIAM.
        """
        curie = self.get_miriam_curie(prefix, identifier)
        if curie is None:
            return None
        return f"{IDENTIFIERS_ORG_URL_PREFIX}{curie}"

    def get_bioportal_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the Bioportal URL for the given CURIE.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A link to the Bioportal page

        >>> from bioregistry import manager
        >>> manager.get_bioportal_iri('chebi', '24431')
        'https://bioportal.bioontology.org/ontologies/CHEBI/?p=classes&conceptid=http://purl.obolibrary.org/obo/CHEBI_24431'
        """
        bioportal_prefix = self.get_mapped_prefix(prefix, "bioportal")
        if bioportal_prefix is None:
            return None
        obo_link = self.get_obofoundry_iri(prefix, identifier)
        if obo_link is not None:
            return f"https://bioportal.bioontology.org/ontologies/{bioportal_prefix}/?p=classes&conceptid={obo_link}"
        # TODO there must be other rules?
        return None

    def get_ols_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the OLS URL if possible."""
        ols_prefix = self.get_mapped_prefix(prefix, "ols")
        obo_iri = self.get_obofoundry_iri(prefix, identifier)
        if ols_prefix is None or obo_iri is None:
            return None
        return f"https://www.ebi.ac.uk/ols/ontologies/{ols_prefix}/terms?iri={obo_iri}"

    def get_formatted_iri(self, metaprefix: str, prefix: str, identifier: str) -> Optional[str]:
        """Get an IRI using the format in the metaregistry.

        :param metaprefix: The metaprefix of the registry in the metaregistry
        :param prefix: A bioregistry prefix (will be mapped to the external one automatically)
        :param identifier: The identifier for the entity
        :returns: An IRI generated from the ``resolver_url`` format string of the registry, if it
            exists.

        >>> from bioregistry import manager
        >>> manager.get_formatted_iri("miriam", "hgnc", "16793")
        'https://identifiers.org/hgnc:16793'
        >>> manager.get_formatted_iri("n2t", "hgnc", "16793")
        'https://n2t.net/hgnc:16793'
        >>> manager.get_formatted_iri("obofoundry", "fbbt", "00007294")
        'http://purl.obolibrary.org/obo/FBbt_00007294'
        >>> manager.get_formatted_iri("scholia", "lipidmaps", "00000052")
        'https://scholia.toolforge.org/lipidmaps/00000052'
        """
        mapped_prefix = self.get_mapped_prefix(prefix, metaprefix)
        registry = self.metaregistry.get(metaprefix)
        if registry is None or mapped_prefix is None:
            return None
        return registry.resolve(mapped_prefix, identifier)

    def get_obofoundry_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the OBO Foundry URL if possible.

        :param prefix: The prefix
        :param identifier: The identifier
        :return: The OBO Foundry URL if the prefix can be mapped to an OBO Foundry entry

        >>> from bioregistry import manager
        >>> manager.get_obofoundry_iri('chebi', '24431')
        'http://purl.obolibrary.org/obo/CHEBI_24431'

        For entries where there's a preferred prefix, it is respected.

        >>> manager.get_obofoundry_iri('fbbt', '00007294')
        'http://purl.obolibrary.org/obo/FBbt_00007294'
        """
        return self.get_formatted_iri("obofoundry", prefix, identifier)

    def get_n2t_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get the name-to-thing URL for the given CURIE.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A IRI string corresponding to the N2T resolve, if the prefix exists and is
            mapped to N2T.

        >>> from bioregistry import manager
        >>> manager.get_n2t_iri("chebi", "24867")
        'https://n2t.net/chebi:24867'
        """
        return self.get_formatted_iri("n2t", prefix, identifier)

    def get_scholia_iri(self, prefix: str, identifier: str) -> Optional[str]:
        """Get a Scholia IRI, if possible.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE
        :return: A link to the Scholia page

        >>> from bioregistry import manager
        >>> manager.get_scholia_iri("pubmed", "1234")
        'https://scholia.toolforge.org/pubmed/1234'

        >>> manager.get_scholia_iri("pdb", "1234")
        None
        """
        return self.get_formatted_iri("scholia", prefix, identifier)

    def get_provider_functions(self) -> Mapping[str, Callable[[str, str], Optional[str]]]:
        """Return a mapping of provider functions."""
        return {
            "default": self.get_default_iri,
            "miriam": self.get_miriam_iri,
            "obofoundry": self.get_obofoundry_iri,
            "ols": self.get_ols_iri,
            "n2t": self.get_n2t_iri,
            "bioportal": self.get_bioportal_iri,
            "scholia": self.get_scholia_iri,
        }

    def get_providers_list(self, prefix: str, identifier: str) -> Sequence[Tuple[str, str]]:
        """Get all providers for the CURIE."""
        rv = []
        for provider, get_url in self.get_provider_functions().items():
            link = get_url(prefix, identifier)
            if link is not None:
                rv.append((provider, link))
        if not rv:
            return rv

        bioregistry_link = self.get_bioregistry_iri(prefix, identifier)
        if not bioregistry_link:
            return rv

        # if a default URL is available, it goes first. otherwise the bioregistry URL goes first.
        rv.insert(1 if rv[0][0] == "default" else 0, ("bioregistry", bioregistry_link))
        return rv

    def get_providers(self, prefix: str, identifier: str) -> Dict[str, str]:
        """Get all providers for the CURIE."""
        return dict(self.get_providers_list(prefix, identifier))

    def get_iri(
        self,
        prefix: str,
        identifier: Optional[str] = None,
        *,
        priority: Optional[Sequence[str]] = None,
        prefix_map: Optional[Mapping[str, str]] = None,
        use_bioregistry_io: bool = True,
    ) -> Optional[str]:
        """Get the best link for the CURIE, if possible.

        :param prefix: The prefix in the CURIE
        :param identifier: The identifier in the CURIE. If identifier is given as None, then this function will
            assume that the first argument (``prefix``) is actually a full CURIE.
        :param priority: A user-defined priority list. In addition to the metaprefixes in the Bioregistry
            corresponding to resources that are resolvers/lookup services, you can also use ``default``
            to correspond to the first-party IRI and ``custom`` to refer to the custom prefix map.
            The default priority list is:

            1. Custom prefix map (``custom``)
            1. First-party IRI (``default``)
            2. Identifiers.org / MIRIAM (``miriam``)
            3. Ontology Lookup Service (``ols``)
            4. OBO PURL (``obofoundry``)
            5. Name-to-Thing (``n2t``)
            6. BioPortal (``bioportal``)
        :param prefix_map: A custom prefix map to go with the ``custom`` key in the priority list
        :param use_bioregistry_io: Should the bioregistry resolution IRI be used? Defaults to true.
        :return: The best possible IRI that can be generated based on the priority list.

        A pre-parse CURIE can be given as the first two arguments
        >>> from bioregistry import manager
        >>> manager.get_iri("chebi", "24867")
        'https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:24867'

        A CURIE can be given directly as a single argument
        >>> manager.get_iri("chebi:24867")
        'https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:24867'

        A priority list can be given
        >>> priority = ["obofoundry", "default", "bioregistry"]
        >>> manager.get_iri("chebi:24867", priority=priority)
        'http://purl.obolibrary.org/obo/CHEBI_24867'

        A custom prefix map can be supplied.
        >>> prefix_map = {"chebi": "https://example.org/chebi/"}
        >>> manager.get_iri("chebi:24867", prefix_map=prefix_map)
        'https://example.org/chebi/24867'
        >>> manager.get_iri("fbbt:00007294")
        'https://flybase.org/cgi-bin/cvreport.pl?id=FBbt:00007294'

        A custom prefix map can be supplied in combination with a priority list
        >>> prefix_map = {"lipidmaps": "https://example.org/lipidmaps/"}
        >>> priority = ["obofoundry", "custom", "default", "bioregistry"]
        >>> manager.get_iri("chebi:24867", prefix_map=prefix_map, priority=priority)
        'http://purl.obolibrary.org/obo/CHEBI_24867'
        >>> manager.get_iri("lipidmaps:1234", prefix_map=prefix_map, priority=priority)
        'https://example.org/lipidmaps/1234'
        """
        if identifier is None:
            _prefix, _identifier = self.parse_curie(prefix)
            if _prefix is None or _identifier is None:
                return None
        else:
            _prefix, _identifier = prefix, identifier

        providers = self.get_providers(_prefix, _identifier)
        if prefix_map and _prefix in prefix_map:
            providers["custom"] = f"{prefix_map[_prefix]}{_identifier}"
        for key in priority or LINK_PRIORITY:
            if not use_bioregistry_io and key == "bioregistry":
                continue
            if key not in providers:
                continue
            rv = providers[key]
            if rv is not None:
                return rv
        return None

    def get_internal_prefix_map(self) -> Mapping[str, str]:
        """Get an internal prefix map for RDF and SSSOM dumps."""
        default_prefixes = {"bioregistry.schema", "bfo"}
        rv = cast(
            Dict[str, str], {prefix: self.get_uri_prefix(prefix) for prefix in default_prefixes}
        )
        for metaprefix, metaresource in self.metaregistry.items():
            uri_prefix = metaresource.get_provider_uri_prefix()
            if metaresource.bioregistry_prefix:
                rv[metaresource.bioregistry_prefix] = uri_prefix
            else:
                rv[metaprefix] = uri_prefix
        return rv


def prepare_prefix_list(prefix_map: Mapping[str, str]) -> List[Tuple[str, str]]:
    """Prepare a priority prefix list from a prefix map."""
    rv = []
    for prefix, uri_prefix in sorted(prefix_map.items(), key=_sort_key):
        rv.append((prefix, uri_prefix))
        if uri_prefix.startswith("https://"):
            rv.append((prefix, "http://" + uri_prefix[8:]))
        elif uri_prefix.startswith("http://"):
            rv.append((prefix, "https://" + uri_prefix[7:]))
    return rv


def _sort_key(kv: Tuple[str, str]) -> int:
    """Return a value appropriate for sorting a pair of prefix/IRI."""
    return -len(kv[0])


#: The default manager for the Bioregistry
manager = Manager()



# -*- coding: utf-8 -*-

"""This script compares what's in each resource."""

import datetime
import itertools as itt
import logging
import math
import random
import sys
import typing
from collections import Counter, defaultdict
from typing import Collection, List, Mapping, Set, Tuple

import click

import bioregistry
from bioregistry import (
    get_contact_email,
    get_description,
    get_example,
    get_external,
    get_homepage,
    get_json_download,
    get_license,
    get_name,
    get_obo_download,
    get_owl_download,
    get_pattern,
    get_uri_format,
    get_version,
    is_deprecated,
    manager,
    read_registry,
)
from bioregistry.constants import DOCS_IMG
from bioregistry.license_standardizer import standardize_license
from bioregistry.schema import Resource

logger = logging.getLogger(__name__)

# see named colors https://matplotlib.org/stable/gallery/color/named_colors.html
BIOREGISTRY_COLOR = "silver"


def _get_has(func, yes: str = "Yes", no: str = "No") -> Counter:
    return Counter(
        no if func(prefix) is None else yes
        for prefix in read_registry()
        if not is_deprecated(prefix)
    )


HAS_WIKIDATA_DATABASE = Counter(
    "No" if key is None else "Yes"
    for key in read_registry()
    if not is_deprecated(key) and "database" in get_external(key, "wikidata")
)


def _get_has_present(func) -> Counter:
    return Counter(x for x in (func(prefix) for prefix in read_registry()) if x)


SINGLE_FIG = (8, 3.5)
TODAY = datetime.datetime.today().strftime("%Y-%m-%d")
WATERMARK_TEXT = f"https://github.com/biopragmatics/bioregistry ({TODAY})"


def _save(fig, name: str, *, svg: bool = True, png: bool = False, eps: bool = False) -> None:
    import matplotlib.pyplot as plt

    path = DOCS_IMG.joinpath(name).with_suffix(".svg")
    click.echo(f"output to {path}")
    fig.tight_layout()
    if svg:
        fig.savefig(path)
    if eps:
        fig.savefig(DOCS_IMG.joinpath(name).with_suffix(".eps"))
    if png:
        fig.savefig(DOCS_IMG.joinpath(name).with_suffix(".png"), dpi=300)
    plt.close(fig)


def _plot_attribute_pies(*, measurements, watermark, ncols: int = 4, keep_ontology: bool = True):
    import matplotlib.pyplot as plt

    if not keep_ontology:
        measurements = [
            (label, counter)
            for label, counter in measurements
            if label not in {"OWL", "JSON", "OBO"}
        ]

    nrows = int(math.ceil(len(measurements) / ncols))
    figsize = (2.75 * ncols, 2.0 * nrows)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    for (label, counter), ax in itt.zip_longest(measurements, axes.ravel(), fillvalue=(None, None)):
        if label is None:
            ax.axis("off")
            continue
        if label == "License Type":
            labels, sizes = zip(*counter.most_common())
            explode = None
        else:
            labels = ("Yes", "No")
            n_yes = counter.get("Yes")
            sizes = (n_yes, len(read_registry()) - n_yes)
            explode = [0.1, 0]
        ax.pie(
            sizes,
            labels=labels,
            autopct="%1.f%%",
            startangle=30,
            explode=explode,
        )
        ax.set_title(label)
    if watermark:
        fig.text(
            0.5,
            0,
            WATERMARK_TEXT,
            fontsize=8,
            color="gray",
            alpha=0.5,
            ha="center",
            va="bottom",
        )
    return fig, axes


REMAPPED_KEY = "x"
REMAPPED_VALUE = "y"


def make_overlaps(keys) -> Mapping[str, Mapping[str, Set[str]]]:
    """Make overlaps ditionary."""
    rv = {}
    for key, _, _, prefixes in keys:
        # Remap bioregistry prefixes to match the external
        #  vocabulary, when possible
        bioregistry_remapped = {
            bioregistry.get_external(br_key, key).get("prefix", br_key)
            for br_key, br_entry in bioregistry.read_registry().items()
        }
        rv[key] = {
            REMAPPED_KEY: bioregistry_remapped,
            REMAPPED_VALUE: prefixes,
        }
    return rv


def _plot_coverage(*, keys, overlaps, watermark, ncols: int = 3):
    import matplotlib.pyplot as plt
    from matplotlib_venn import venn2

    nrows = int(math.ceil(len(keys) / ncols))
    figsize = (3.25 * ncols, 2.0 * nrows)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)

    for key, ax in itt.zip_longest(keys, axes.ravel()):
        if key is None:
            ax.axis("off")
            continue
        key, label, color, prefixes = key
        bioregistry_remapped = overlaps[key][REMAPPED_KEY]
        venn2(
            subsets=(bioregistry_remapped, prefixes),
            set_labels=("Bioregistry", label),
            set_colors=(BIOREGISTRY_COLOR, color),
            ax=ax,
        )
    if watermark:
        fig.text(
            0.5,
            0,
            WATERMARK_TEXT,
            fontsize=8,
            color="gray",
            alpha=0.5,
            ha="center",
            va="bottom",
        )
    return fig, axes


def _plot_external_overlap(*, keys, watermark, ncols: int = 4):
    import matplotlib.pyplot as plt
    from matplotlib_venn import venn2

    pairs = list(itt.combinations(keys, r=2))
    nrows = int(math.ceil(len(pairs) / ncols))
    figsize = (3 * ncols, 2.5 * nrows)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    for pair, ax in itt.zip_longest(pairs, axes.ravel()):
        if pair is None:
            ax.axis("off")
            continue
        (l_key, l_label, l_color, l_prefixes), (r_key, r_label, r_color, r_prefixes) = pair
        # Remap external vocabularies to bioregistry
        #  prefixes, when possible
        l_prefixes = _remap(key=l_key, prefixes=l_prefixes)
        r_prefixes = _remap(key=r_key, prefixes=r_prefixes)
        venn2(
            subsets=(l_prefixes, r_prefixes),
            set_labels=(l_label, r_label),
            set_colors=(l_color, r_color),
            ax=ax,
        )
    if watermark:
        fig.text(
            0.5,
            0,
            WATERMARK_TEXT,  # transform=plt.gca().transAxes,
            fontsize=14,
            color="gray",
            alpha=0.5,
            ha="center",
            va="bottom",
        )
    return fig, axes


def get_getters():
    """Get getter functions, which requires alignment dependencies."""
    try:
        from bioregistry.external import GETTERS
    except ImportError:
        click.secho(
            "Could not import alignment dependencies."
            " Install bioregistry again with `pip install bioregistry[align]`.",
            fg="red",
        )
        return sys.exit(1)
    else:
        return GETTERS


def get_keys() -> List[Tuple[str, str, str, Set[str]]]:
    """Get keys for plots."""
    getters = get_getters()
    try:
        import seaborn as sns
    except ImportError:
        raise
    else:
        palette = sns.color_palette("Paired", len(getters))
    return [
        (metaprefix, label, color, set(func(force_download=False)))
        for (metaprefix, label, func), color in zip(getters, palette)
    ]


@click.command()
@click.option("--paper", is_flag=True)
def compare(paper: bool):  # noqa:C901
    """Compare the registries."""
    paper = True
    random.seed(0)
    try:
        import matplotlib.pyplot as plt
        import pandas as pd
        import seaborn as sns
    except ImportError:
        click.secho(
            "Could not import matplotlib dependencies."
            " Install bioregistry again with `pip install bioregistry[charts]`.",
            fg="red",
        )
        return sys.exit(1)

    keys = get_keys()
    overlaps = make_overlaps(keys)

    # This should make SVG output deterministic
    # See https://matplotlib.org/3.1.0/users/prev_whats_new/whats_new_2.0.0.html#added-svg-hashsalt-key-to-rcparams
    plt.rcParams["svg.hashsalt"] = "saltyregistry"

    watermark = True
    sns.set_style("white")

    ###############################################
    # What kinds of licenses are resources using? #
    ###############################################
    licenses, conflicts, obo_has_license, ols_has_license = _get_license_and_conflicts()

    # How many times does the license appear in OLS / OBO Foundry
    # fig, ax = plt.subplots(figsize=SINGLE_FIG)
    # venn2(
    #     subsets=(obo_has_license, ols_has_license),
    #     set_labels=("OBO", "OLS"),
    #     set_colors=("red", "green"),
    #     ax=ax,
    # )
    # if watermark:
    #     ax.text(
    #         0.5,
    #         -0.1,
    #         WATERMARK_TEXT,
    #         transform=plt.gca().transAxes,
    #         fontsize=10,
    #         color="gray",
    #         alpha=0.5,
    #         ha="center",
    #         va="bottom",
    #     )
    # _save(fig, name="license_coverage", eps=paper)

    fig, ax = plt.subplots(figsize=SINGLE_FIG)
    licenses_counter: typing.Counter[str] = Counter(licenses)
    licenses_mapped = [
        "None" if license_ is None else license_ if licenses_counter[license_] > 30 else "Other"
        for license_ in licenses
    ]
    licenses_mapped_counter = Counter(licenses_mapped)
    licenses_mapped_order = [license_ for license_, _ in licenses_mapped_counter.most_common()]
    sns.countplot(x=licenses_mapped, ax=ax, order=licenses_mapped_order)
    ax.set_xlabel("License")
    ax.set_ylabel("Count")
    ax.set_yscale("log")
    if watermark:
        fig.text(
            1.0,
            0.5,
            WATERMARK_TEXT,
            fontsize=8,
            color="gray",
            alpha=0.5,
            ha="right",
            va="center",
            rotation=90,
        )
    _save(fig, name="licenses", eps=paper)

    ##############################################
    # How many entries have version information? #
    ##############################################
    measurements = [
        ("Name", _get_has(get_name)),
        ("Homepage", _get_has(get_homepage)),
        ("Description", _get_has(get_description)),
        ("Example", _get_has(get_example)),
        ("Pattern", _get_has(get_pattern)),
        ("Provider", _get_has(get_uri_format)),
        ("License", _get_has(get_license)),
        ("License Type", _get_has_present(get_license)),
        ("Version", _get_has(get_version)),
        ("Contact Email", _get_has(get_contact_email)),
        ("Wikidata Database", HAS_WIKIDATA_DATABASE),
        ("OBO", _get_has(get_obo_download)),
        ("OWL", _get_has(get_owl_download)),
        ("JSON", _get_has(get_json_download)),
    ]

    fig, axes = _plot_attribute_pies(measurements=measurements, watermark=watermark)
    _save(fig, "has_attribute", eps=paper)

    # Slightly reorganized for the paper
    if paper:
        fig, axes = _plot_attribute_pies(
            measurements=measurements, watermark=watermark, keep_ontology=False
        )
        _save(fig, "paper_figure_3", png=True, eps=True)

    # -------------------------------------------------------------------- #

    ############################################################
    # How well does the Bioregistry cover the other resources? #
    ############################################################
    fig, axes = _plot_coverage(keys=keys, overlaps=overlaps, watermark=watermark)
    _save(fig, name="bioregistry_coverage", eps=paper)
    plot_coverage_bar_abridged(overlaps=overlaps, paper=paper)
    plot_coverage_bar(overlaps=overlaps, paper=True)

    ######################################################
    # What's the overlap between each pair of resources? #
    ######################################################
    fig, axes = _plot_external_overlap(keys=keys, watermark=watermark)
    _save(fig, name="external_overlap", eps=paper)

    ##############################################
    # Histogram of how many xrefs each entry has #
    ##############################################
    xref_counts = [
        sum(0 < len(entry.get_external(key)) for key, *_ in keys)
        for entry in read_registry().values()
    ]
    fig, ax = plt.subplots(figsize=SINGLE_FIG)
    xrefs_counter: typing.Counter[int] = Counter(xref_counts)

    n_mappable_metaprefixes = len(
        {
            metaprefix
            for entry in read_registry().values()
            for metaprefix in (entry.get_mappings() or {})
        }
    )
    zero_pad_count = 0  # how many columns left from the end should it go
    for i in range(n_mappable_metaprefixes):
        if i not in xrefs_counter:
            zero_pad_count += 1
            xrefs_counter[i] = 0

    xrefs_df = pd.DataFrame(sorted(xrefs_counter.items()), columns=["frequency", "count"])
    palette = sns.color_palette("tab10")
    xrefs_colors = [palette[2]] + ([palette[1]] * (len(xrefs_df.index) - 1))
    sns.barplot(
        data=xrefs_df,
        x="frequency",
        y="count",
        ci=None,
        palette=xrefs_colors,
        alpha=1.0,
        ax=ax,
    )
    # There should only be one container here
    _labels = xrefs_df["count"].to_list()
    _labels[0] = f"{_labels[0]}\nNovel"
    for i in ax.containers:
        ax.bar_label(i, _labels)
    ax.set_xlabel("Number Cross-Registry Mappings")
    ax.set_ylabel("Number Prefixes")
    ax.set_yscale("log")
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    h = 15  # how high should the text go
    x1, _y1 = ax.patches[-zero_pad_count].get_xy()
    x2, _y2 = ax.patches[-1].get_xy()
    ax.text(
        x1,
        h + 1,
        "No prefixes are available\nin $\\it{all}$ mappable external\nregistries",
        horizontalalignment="center",
        verticalalignment="bottom",
        fontdict=dict(fontsize=12),
    )
    ax.arrow(x1, h, x2 - x1, 2 - h, head_width=0.3, head_length=0.2, fc="k", ec="k")
    if watermark:
        fig.text(
            1.0,
            0.5,
            WATERMARK_TEXT,
            fontsize=8,
            color="gray",
            alpha=0.5,
            ha="right",
            va="center",
            rotation=90,
        )

    offset = 0.6
    ax.set_xlim([-offset, len(ax.patches) - (1 + offset)])
    _save(fig, name="xrefs", eps=paper, png=paper)

    ##################################################
    # Histogram of how many providers each entry has #
    ##################################################
    provider_counts = [_count_providers(resource) for resource in read_registry().values()]
    fig, ax = plt.subplots(figsize=SINGLE_FIG)
    sns.barplot(
        data=sorted(Counter(provider_counts).items()), ci=None, color="blue", alpha=0.4, ax=ax
    )
    ax.set_xlabel("Number Providers")
    ax.set_ylabel("Count")
    ax.set_yscale("log")
    if watermark:
        fig.text(
            1.0,
            0.5,
            WATERMARK_TEXT,
            fontsize=8,
            color="gray",
            alpha=0.5,
            ha="right",
            va="center",
            rotation=90,
        )
    _save(fig, name="providers", eps=paper)

    ########################################
    # Regular expression complexity report #
    ########################################
    g = sns.displot(x=get_regex_complexities(), log_scale=2, height=3, aspect=4 / 3)
    g.set(xlabel="Regular Expression Complexity")
    _save(g.figure, name="regex_report", eps=paper)


def _count_providers(resource: Resource) -> int:
    rv = 0
    if resource.get_uri_prefix():
        rv += 1
    rv += len(resource.get_extra_providers())
    return rv


def _get_license_and_conflicts():
    licenses = []
    conflicts = set()
    obo_has_license, ols_has_license = set(), set()
    for key in read_registry():
        obo_license = standardize_license(get_external(key, "obofoundry").get("license"))
        if obo_license:
            obo_has_license.add(key)

        ols_license = standardize_license(get_external(key, "ols").get("license"))
        if ols_license:
            ols_has_license.add(key)

        if not obo_license and not ols_license:
            licenses.append("None")
        if obo_license and not ols_license:
            licenses.append(obo_license)
        elif not obo_license and ols_license:
            licenses.append(ols_license)
        elif obo_license == ols_license:
            licenses.append(obo_license)
        else:  # different licenses!
            licenses.append(ols_license)
            licenses.append(obo_license)
            conflicts.add(key)
            # logger.warning(f"[{key}] Conflicting licenses- {obo_license} and {ols_license}")
            continue
    return licenses, conflicts, obo_has_license, ols_has_license


def _remap(*, key: str, prefixes: Collection[str]) -> Set[str]:
    br_external_to = {}
    for br_id, resource in read_registry().items():
        _k = (resource.dict().get(key) or {}).get("prefix")
        if _k:
            br_external_to[_k] = br_id

    return {br_external_to.get(prefix, prefix) for prefix in prefixes}


def get_regex_complexities() -> Collection[float]:
    """Get a list of regular expression complexities."""
    rows = []
    for prefix in manager.registry:
        pattern = manager.get_pattern(prefix)
        if pattern is None:
            continue
        # Consider alternate complexity estimates
        rows.append(float(len(pattern)))
    return sorted(rows)


def plot_coverage_bar_abridged(*, overlaps, paper: bool = False):
    """Plot and save the abridged coverage bar chart."""
    import matplotlib.pyplot as plt
    import pandas as pd
    import seaborn as sns

    sns.set_style("white")

    rows = []
    for metaprefix, data in overlaps.items():
        br, external = data[REMAPPED_KEY], data[REMAPPED_VALUE]
        rows.append(
            (
                bioregistry.get_registry_short_name(metaprefix),
                len(external - br),
                len(br.intersection(external)),
            )
        )
    rows = sorted(rows, key=lambda row: sum(row[1:]), reverse=True)

    df2 = pd.DataFrame(rows, columns=["metaprefix", "external_only", "intersection"])
    df2.set_index("metaprefix", inplace=True)

    fig, ax = plt.subplots(1, 1, figsize=(10, 7))
    df2.plot(
        kind="barh",
        stacked=True,
        ax=ax,
        width=0.85,
        fontsize=14,
        grid=False,
    )
    ax.grid(False)
    ax.set_ylabel("")
    ax.set_xticks([])
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.tick_params(
        top=False, bottom=False, left=False, right=False, labelleft=True, labelbottom=False
    )

    dd = defaultdict(list)
    for p in ax.patches:
        width, height = p.get_width(), p.get_height()
        x, y = p.get_xy()
        dd[y, height].append((width, x))
        if width < 20:
            continue
        ax.text(
            x + width / 2,
            y + height / 2,
            f"{int(width):,}",
            horizontalalignment="center",
            verticalalignment="center",
            fontdict=dict(weight="bold", color="white", fontsize=12),
        )

    for (y, height), values in dd.items():
        width_total = sum(int(w) for w, _ in values)
        percentage = values[-1][0] / width_total
        width, x = max(values, key=lambda item: item[1])
        ax.text(
            width + x + 20,
            y + height / 2,
            f"{percentage:.1%} coverage",
            fontdict=dict(weight="normal", color="black", fontsize=12),
            verticalalignment="center",
        )

    ax.get_legend().remove()
    plt.tight_layout()
    _save(fig, name="bioregistry_coverage_bar_short", eps=paper)


def plot_coverage_bar(*, overlaps, paper: bool = False):
    """Plot and save the coverage bar chart."""
    import matplotlib.pyplot as plt
    import pandas as pd
    import seaborn as sns

    sns.set_style("white")

    rows_1 = []
    for metaprefix, data in overlaps.items():
        br, external = data[REMAPPED_KEY], data[REMAPPED_VALUE]
        rows_1.append(
            (
                bioregistry.get_registry_short_name(metaprefix),
                len(external - br),
                len(br.intersection(external)),
                len(br - external),
            )
        )
    rows_1 = sorted(rows_1, key=lambda row: sum(row[1:]), reverse=True)

    df1 = pd.DataFrame(
        rows_1, columns=["metaprefix", "external_only", "intersection", "bioregistry_only"]
    )
    df1.set_index("metaprefix", inplace=True)

    fig, ax = plt.subplots(1, 1, figsize=(10, 7))
    df1.plot(
        kind="barh",
        stacked=True,
        ax=ax,
        width=0.85,
        fontsize=14,
        grid=False,
    )
    ax.set_ylabel("")
    ax.set_xticks([])
    ax.grid(False)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.tick_params(
        top=False, bottom=False, left=False, right=False, labelleft=True, labelbottom=False
    )

    dd = defaultdict(list)
    for p in ax.patches:
        width, height = p.get_width(), p.get_height()
        x, y = p.get_xy()
        dd[y, height].append((width, x))
        if width < 40:
            continue
        ax.text(
            x + width / 2,
            y + height / 2,
            f"{int(width):,}",
            horizontalalignment="center",
            verticalalignment="center",
            fontdict=dict(weight="bold", color="white", fontsize=12),
        )

    for (y, height), values in dd.items():
        width_total = sum(int(w) for w, _ in values)
        without_br = sum(int(w) for w, _ in values[:-1])
        increase = (width_total - without_br) / without_br
        width, x = max(values, key=lambda item: item[1])
        ax.text(
            width + x + 20,
            y + height / 2,
            f"{int(width_total):,} (+{increase:,.0%})",
            fontdict=dict(weight="normal", color="black", fontsize=12),
            verticalalignment="center",
        )

    for label in ax.get_yticklabels():
        label.set_fontweight("bold")

    ax.get_legend().remove()
    plt.tight_layout()
    _save(fig, name="bioregistry_coverage_bar", eps=paper, png=True)


if __name__ == "__main__":
    compare()



# -*- coding: utf-8 -*-

"""Download registry information from the Life Science Registry (LSR), which powers Prefix Commons.

.. seealso::

    - http://tinyurl.com/lsregistry which should expand to
      https://docs.google.com/spreadsheets/d/1cDGJcRteb9F5-jbw7Q7np0kk4hfWhdBHNYRIg3LXDrs/edit#gid=0
"""

import json
from typing import Any, Dict

from pystow.utils import download

from bioregistry.constants import EXTERNAL

__all__ = [
    "get_prefixcommons",
]

DIRECTORY = EXTERNAL / "prefixcommons"
DIRECTORY.mkdir(exist_ok=True, parents=True)
RAW_PATH = DIRECTORY / "raw.tsv"
PROCESSED_PATH = DIRECTORY / "processed.json"
GOOGLE_DOCUMENT_ID = "1c4DmQqTGS4ZvJU_Oq2MFnLk-3UUND6pWhuMoP8jgZhg"
URL = f"https://docs.google.com/spreadsheets/d/{GOOGLE_DOCUMENT_ID}/export?format=tsv&gid=0"
COLUMNS = [
    "prefix",  # "Preferred Prefix",
    "Alt-prefix",
    "Provider Base URI",
    "Alternative Base URI",
    "MIRIAM",
    "BiodbcoreID",
    "bioportal",  # "BioPortal Ontology ID",
    "miriam",  # "identifiers.org",
    "Abbreviation",
    "name",  # originally: Title,
    "description",  # "Description",
    "pubmed_ids",  # "PubMed ID"
    "Organization",
    "Type (warehouse, dataset or terminology)",
    "keywords",
    "homepage",  # "Homepage",
    "Functional?",
    "sub-namespace in dataset",
    "part of collection",
    "License URL",
    "License Text",
    "Rights",
    "pattern",  # "ID regex",
    "example",  # "ExampleID",
    "uri_format",  # "Provider HTML URL",
    "",
    "MIRIAM checked",
    "MIRIAM curator notes",
    "MIRIAM coverage",
    "updates",
    "year last accessible",
    "wayback url",
    "last updated",
    "last updated by",
    "last updated by (orcid)",
]
KEEP = {
    "prefix",
    "bioportal",
    "miriam",
    "name",
    "description",
    "pubmed_ids",
    "keywords",
    "homepage",
    "pattern",
    "example",
    "uri_format",
}


def get_prefixcommons(force_download: bool = False):
    """Get the Life Science Registry."""
    if PROCESSED_PATH.exists() and not force_download:
        with PROCESSED_PATH.open() as file:
            return json.load(file)

    download(url=URL, path=RAW_PATH, force=True)
    rows = {}
    with RAW_PATH.open() as file:
        lines = iter(file)
        next(lines)  # throw away header
        for line in lines:
            prefix, data = _process_row(line)
            if prefix and data:
                rows[prefix] = data

    PROCESSED_PATH.write_text(json.dumps(rows, sort_keys=True, indent=2))
    return rows


def _process_row(line: str):
    cells = line.strip().split("\t")
    cells_processed = [None if cell in {"N/A"} else cell for cell in cells]
    rv: Dict[str, Any] = {
        key: value.strip()
        for key, value in zip(COLUMNS, cells_processed)
        if key and value and key in KEEP
    }
    for key in ["name", "description", "example", "pattern"]:
        if not rv.get(key):
            return None, None

    for key in ["keywords", "pubmed_ids"]:
        values = rv.get(key)
        if values:
            rv[key] = [value.strip() for value in values.split(",")]

    uri_format = rv.get("uri_format")
    if uri_format:
        rv["uri_format"] = uri_format.replace("$id", "$1")

    pattern = rv.get("pattern")
    if pattern:
        if not pattern.startswith("^"):
            pattern = f"^{pattern}"
        if not pattern.endswith("$"):
            pattern = f"{pattern}$"
        rv["pattern"] = pattern

    return cells[0], rv


if __name__ == "__main__":
    print(len(get_prefixcommons(force_download=True)))  # noqa:T201



"""Utilities."""

import logging
from dataclasses import asdict, is_dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, List, Mapping, Optional, Union

import click
import requests
from pydantic import BaseModel
from pydantic.json import ENCODERS_BY_TYPE
from pystow.utils import get_hashes

from .constants import (
    BIOREGISTRY_PATH,
    COLLECTIONS_YAML_PATH,
    METAREGISTRY_YAML_PATH,
    REGISTRY_YAML_PATH,
)

logger = logging.getLogger(__name__)

#: Wikidata SPARQL endpoint. See https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service#Interfacing
WIKIDATA_ENDPOINT = "https://query.wikidata.org/bigdata/namespace/wdq/sparql"


def secho(s, fg="cyan", bold=True, **kwargs):
    """Wrap :func:`click.secho`."""
    click.echo(
        f'[{datetime.now().strftime("%H:%M:%S")}] ' + click.style(s, fg=fg, bold=bold, **kwargs)
    )


def removeprefix(s: Optional[str], prefix: str) -> Optional[str]:
    """Remove the prefix from the string."""
    if s is None:
        return None
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


def removesuffix(s: Optional[str], suffix: str) -> Optional[str]:
    """Remove the prefix from the string."""
    if s is None:
        return None
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


def query_wikidata(sparql: str) -> List[Mapping[str, Any]]:
    """Query Wikidata's sparql service.

    :param sparql: A SPARQL query string
    :return: A list of bindings
    """
    logger.debug("running query: %s", sparql)
    res = requests.get(WIKIDATA_ENDPOINT, params={"query": sparql, "format": "json"})
    res.raise_for_status()
    res_json = res.json()
    return res_json["results"]["bindings"]


def extended_encoder(obj: Any) -> Any:
    """Encode objects similarly to :func:`pydantic.json.pydantic_encoder`."""
    if isinstance(obj, BaseModel):
        return obj.dict(exclude_none=True)
    elif is_dataclass(obj):
        return asdict(obj)

    # Check the class type and its superclasses for a matching encoder
    for base in obj.__class__.__mro__[:-1]:
        try:
            encoder = ENCODERS_BY_TYPE[base]
        except KeyError:
            continue
        return encoder(obj)
    else:  # We have exited the for loop without finding a suitable encoder
        raise TypeError(f"Object of type '{obj.__class__.__name__}' is not JSON serializable")


class NormDict(dict):
    """A dictionary that supports lexical normalization of keys on setting and getting."""

    def __setitem__(self, key: str, value: str) -> None:
        """Set an item from the dictionary after lexically normalizing it."""
        norm_key = _norm(key)
        if value is None:
            raise ValueError(f"Tried to add empty value for {key}/{norm_key}")
        if norm_key in self and self[norm_key] != value:
            raise KeyError(
                f"Tried to add {norm_key}/{value} when already had {norm_key}/{self[norm_key]}"
            )
        super().__setitem__(norm_key, value)

    def __getitem__(self, item: str) -> str:
        """Get an item from the dictionary after lexically normalizing it."""
        return super().__getitem__(_norm(item))

    def __contains__(self, item) -> bool:
        """Check if an item is in the dictionary after lexically normalizing it."""
        return super().__contains__(_norm(item))

    def get(self, key: str, default=None) -> str:
        """Get an item from the dictionary after lexically normalizing it."""
        return super().get(_norm(key), default)


def _norm(s: str) -> str:
    """Normalize a string for dictionary key usage."""
    rv = s.casefold().lower()
    for x in " -_./":
        rv = rv.replace(x, "")
    return rv


def norm(s: str) -> str:
    """Normalize a string for dictionary key usage."""
    rv = s.lower()
    for x in " .-":
        rv = rv.replace(x, "")
    return rv


def curie_to_str(prefix: str, identifier: str) -> str:
    """Combine a prefix and identifier into a CURIE string."""
    return f"{prefix}:{identifier}"


def get_hexdigests(alg: str = "sha256") -> Mapping[str, str]:
    """Get hex digests."""
    return {
        path.as_posix(): _get_hexdigest(path, alg=alg)
        for path in (
            BIOREGISTRY_PATH,
            REGISTRY_YAML_PATH,
            METAREGISTRY_YAML_PATH,
            COLLECTIONS_YAML_PATH,
        )
    }


def _get_hexdigest(path: Union[str, Path], alg: str = "sha256") -> str:
    hashes = get_hashes(path, [alg])
    return hashes[alg].hexdigest()



# -*- coding: utf-8 -*-

"""Command line interface for the bioregistry."""

import sys

import click
from more_click import make_web_command

from .compare import compare
from .export.cli import export
from .lint import lint
from .utils import get_hexdigests, secho
from .version import VERSION


@click.group()
@click.version_option(version=VERSION)
def main():
    """Run the Bioregistry CLI."""


@click.command()
def download():
    """Download/update the external entries in the Bioregistry."""
    try:
        from .external import GETTERS
    except ImportError:
        click.secho(
            "Could not import alignment dependencies."
            " Install bioregistry again with `pip install bioregistry[align]`.",
            fg="red",
        )
        return sys.exit(1)

    for _, name, getter in GETTERS:
        secho(f"Downloading {name}")
        getter(force_download=True)


@main.command()
@click.option("--skip-fairsharing", is_flag=True)
def align(skip_fairsharing: bool):
    """Align all external registries."""
    try:
        from .align import ALIGNERS
    except ImportError:
        click.secho(
            "Could not import alignment dependencies."
            " Install bioregistry again with `pip install bioregistry[align]`.",
            fg="red",
        )
        return sys.exit(1)

    pre_digests = get_hexdigests()
    aligners = [a for a in ALIGNERS if a.key != "fairsharing"] if skip_fairsharing else ALIGNERS

    for aligner_cls in aligners:
        secho(f"Aligning {aligner_cls.key}")
        try:
            aligner_cls.align()
        except IOError as e:
            secho(f"Failed to align {aligner_cls.key}: {e}", fg="red")

    if pre_digests != get_hexdigests():
        secho("Alignment created updates", fg="green")
        click.echo("::set-output name=BR_UPDATED::true")


main.add_command(lint)
main.add_command(compare)
main.add_command(export)
main.add_command(make_web_command("bioregistry.app.wsgi:app"))


@main.command()
@click.pass_context
def update(ctx: click.Context):
    """Update the Bioregistry."""
    ctx.invoke(align)
    ctx.invoke(lint)
    ctx.invoke(export)
    ctx.invoke(compare)

    try:
        from . import upload_ndex
    except ImportError:
        click.secho("Could not import ndex")
    else:
        ctx.invoke(upload_ndex.main)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Export the Bioregistry to RDF."""

import logging
from io import BytesIO
from typing import Any, Callable, List, Optional, Tuple, Union, cast

import click
import rdflib
from rdflib import Literal, Namespace
from rdflib.namespace import DC, DCTERMS, FOAF, RDF, RDFS, SKOS, XSD
from rdflib.term import Node, URIRef

import bioregistry
from bioregistry import manager, read_collections, read_metaregistry, read_registry
from bioregistry.constants import (
    RDF_JSONLD_PATH,
    RDF_NT_PATH,
    RDF_TURTLE_PATH,
    SCHEMA_JSONLD_PATH,
    SCHEMA_NT_PATH,
    SCHEMA_TURTLE_PATH,
)
from bioregistry.export.sssom_export import CURIE_MAP
from bioregistry.schema.constants import (
    bioregistry_collection,
    bioregistry_metaresource,
    bioregistry_resource,
    bioregistry_schema,
    get_schema_rdf,
    orcid,
)
from bioregistry.schema.struct import Collection, Registry

logger = logging.getLogger(__name__)

NAMESPACES = {_ns: Namespace(_uri) for _ns, _uri in CURIE_MAP.items()}
NAMESPACE_WARNINGS = set()


@click.command()
def export_rdf():
    """Export RDF."""
    schema_rdf = get_schema_rdf()
    schema_rdf.serialize(SCHEMA_TURTLE_PATH.as_posix(), format="turtle")
    schema_rdf.serialize(SCHEMA_NT_PATH.as_posix(), format="nt", encoding="utf-8")
    schema_rdf.serialize(
        SCHEMA_JSONLD_PATH.as_posix(),
        format="json-ld",
        context={
            "@language": "en",
            **dict(schema_rdf.namespaces()),
        },
        sort_keys=True,
        ensure_ascii=False,
    )

    graph = get_full_rdf() + schema_rdf
    graph.serialize(RDF_TURTLE_PATH.as_posix(), format="turtle")
    graph.serialize(RDF_NT_PATH.as_posix(), format="nt", encoding="utf-8")
    # Currently getting an issue with not being able to shorten URIs
    # graph.serialize(os.path.join(DOCS_DATA, "bioregistry.xml"), format="xml")

    context = {
        "@language": "en",
        **dict(graph.namespaces()),
    }
    graph.serialize(
        RDF_JSONLD_PATH.as_posix(),
        format="json-ld",
        context=context,
        sort_keys=True,
        ensure_ascii=False,
    )


def _graph() -> rdflib.Graph:
    graph = rdflib.Graph()
    _bind(graph)
    return graph


def _bind(graph: rdflib.Graph) -> None:
    graph.namespace_manager.bind("bioregistry", bioregistry_resource)
    graph.namespace_manager.bind("bioregistry.metaresource", bioregistry_metaresource)
    graph.namespace_manager.bind("bioregistry.collection", bioregistry_collection)
    graph.namespace_manager.bind("bioregistry.schema", bioregistry_schema)
    graph.namespace_manager.bind("orcid", orcid)
    graph.namespace_manager.bind("foaf", FOAF)
    graph.namespace_manager.bind("dc", DC)
    graph.namespace_manager.bind("dcterms", DCTERMS)
    for key, value in CURIE_MAP.items():
        graph.namespace_manager.bind(key, value)


def get_full_rdf() -> rdflib.Graph:
    """Get a combine RDF graph representing the Bioregistry using :mod:`rdflib`."""
    graph = _graph()
    _add_metaresources(graph=graph)
    _add_collections(graph=graph)
    _add_resources(graph=graph)
    return graph


def collection_to_rdf_str(data: Union[str, Collection], fmt: Optional[str] = None) -> str:
    """Get a collection as an RDF string."""
    if isinstance(data, str):
        data = bioregistry.get_collection(data)  # type: ignore
        if data is None:
            raise KeyError
    graph, _ = _add_collection(cast(Collection, data))
    return _graph_str(graph, fmt=fmt)


def metaresource_to_rdf_str(data: Union[str, Registry], fmt: Optional[str] = None) -> str:
    """Get a collection as an RDF string."""
    if isinstance(data, str):
        data = bioregistry.get_registry(data)  # type: ignore
        if data is None:
            raise KeyError
    graph, _ = _add_metaresource(cast(Registry, data))
    return _graph_str(graph, fmt=fmt)


def resource_to_rdf_str(data, fmt: Optional[str] = None) -> str:
    """Get a collection as an RDF string."""
    if isinstance(data, str):
        data = {"prefix": data, **bioregistry.get_resource(data).dict()}  # type: ignore
    graph = _add_resource(data)
    return _graph_str(graph, fmt=fmt)


def _graph_str(graph: rdflib.Graph, fmt: Optional[str] = None) -> str:
    stream = BytesIO()
    graph.serialize(stream, format=fmt or "turtle")
    return stream.getvalue().decode("utf8")


def _add_metaresources(*, graph: Optional[rdflib.Graph] = None) -> rdflib.Graph:
    if graph is None:
        graph = _graph()
    for data in read_metaregistry().values():
        _add_metaresource(graph=graph, data=data)
    return graph


def _add_collections(*, graph: Optional[rdflib.Graph] = None) -> rdflib.Graph:
    if graph is None:
        graph = _graph()
    for collection in read_collections().values():
        _add_collection(graph=graph, data=collection)
    return graph


def _add_resources(*, graph: Optional[rdflib.Graph] = None) -> rdflib.Graph:
    if graph is None:
        graph = _graph()
    for prefix, data in read_registry().items():
        _add_resource(graph=graph, data={"prefix": prefix, **data.dict()})
    return graph


def _add_collection(
    data: Collection, *, graph: Optional[rdflib.Graph] = None
) -> Tuple[rdflib.Graph, Node]:
    if graph is None:
        graph = _graph()
    node = data.add_triples(graph)
    return graph, node


def _add_metaresource(
    data: Registry, *, graph: Optional[rdflib.Graph] = None
) -> Tuple[rdflib.Graph, Node]:
    if graph is None:
        graph = _graph()
    node = data.add_triples(graph)
    return graph, node


RESOURCE_FUNCTIONS: List[Tuple[Union[str, URIRef], Callable[[str], Any], URIRef]] = [
    ("0000008", bioregistry.get_pattern, XSD.string),
    ("0000006", bioregistry.get_uri_format, XSD.string),
    ("0000005", bioregistry.get_example, XSD.string),
    ("0000012", bioregistry.is_deprecated, XSD.boolean),
    (DC.description, bioregistry.get_description, XSD.string),
    (FOAF.homepage, bioregistry.get_homepage, XSD.string),
]


def _add_resource(data, *, graph: Optional[rdflib.Graph] = None) -> rdflib.Graph:
    if graph is None:
        graph = _graph()
    prefix = data["prefix"]
    resource = bioregistry.get_resource(prefix)
    if resource is None:
        logger.warning("Could not look up prefix: %s", prefix)
        return graph
    node = cast(URIRef, bioregistry_resource[prefix])
    graph.add((node, RDF.type, bioregistry_schema["0000001"]))
    graph.add((node, RDFS.label, Literal(bioregistry.get_name(prefix))))
    graph.add((node, DCTERMS.isPartOf, bioregistry_metaresource["bioregistry"]))
    graph.add((bioregistry_metaresource["bioregistry"], DCTERMS.hasPart, node))

    for predicate, func, datatype in RESOURCE_FUNCTIONS:
        value = func(prefix)
        if not isinstance(predicate, URIRef):
            predicate = bioregistry_schema[predicate]
        if value is not None:
            graph.add((node, predicate, Literal(value, datatype=datatype)))

    download = data.get("download")
    if download:
        graph.add((node, bioregistry_schema["0000010"], Literal(download)))

    # Ontological relationships

    for depends_on in bioregistry.get_depends_on(prefix) or []:
        graph.add((node, bioregistry_schema["0000017"], bioregistry_resource[depends_on]))

    for appears_in in bioregistry.get_appears_in(prefix) or []:
        graph.add((node, bioregistry_schema["0000018"], bioregistry_resource[appears_in]))

    part_of = bioregistry.get_part_of(prefix)
    if part_of:
        graph.add((node, DCTERMS.isPartOf, bioregistry_resource[part_of]))
        graph.add((bioregistry_resource[part_of], DCTERMS.hasPart, node))

    provides = bioregistry.get_provides_for(prefix)
    if provides:
        graph.add((node, bioregistry_schema["0000011"], bioregistry_resource[provides]))

    canonical = bioregistry.get_has_canonical(prefix)
    if canonical:
        graph.add((node, bioregistry_schema["0000016"], bioregistry_resource[canonical]))

    contact = bioregistry.get_contact(prefix)
    if contact is not None:
        contact_node = contact.add_triples(graph)
        graph.add((node, bioregistry_schema["0000019"], contact_node))
    if resource.reviewer is not None and resource.reviewer.orcid:
        reviewer_node = resource.reviewer.add_triples(graph)
        graph.add((node, bioregistry_schema["0000021"], reviewer_node))
    if resource.contributor is not None and resource.contributor.orcid:
        contributor_node = resource.contributor.add_triples(graph)
        graph.add((contributor_node, DCTERMS.contributor, node))

    mappings = bioregistry.get_mappings(prefix)
    for metaprefix, metaidentifier in (mappings or {}).items():
        metaresource = manager.metaregistry[metaprefix]
        if metaprefix not in NAMESPACES and metaresource.bioregistry_prefix in NAMESPACES:
            metaprefix = metaresource.bioregistry_prefix
        if metaprefix not in NAMESPACES:
            if metaprefix not in NAMESPACE_WARNINGS:
                logger.warning(f"can not find prefix-uri pair for {metaprefix}")
                NAMESPACE_WARNINGS.add(metaprefix)
            continue
        graph.add((node, SKOS.exactMatch, NAMESPACES[metaprefix][metaidentifier]))
        graph.add(
            (
                NAMESPACES[metaprefix][metaidentifier],
                DCTERMS.isPartOf,
                bioregistry_metaresource[metaresource.prefix],
            )
        )
        graph.add(
            (
                bioregistry_metaresource[metaresource.prefix],
                DCTERMS.hasPart,
                NAMESPACES[metaprefix][metaidentifier],
            )
        )

    return graph


if __name__ == "__main__":
    export_rdf()



"""A script for renaming a metaprefix in the Bioregistry."""

import json

import click

from bioregistry import Resource, write_registry
from bioregistry.constants import BIOREGISTRY_PATH


@click.command()
@click.argument("old_metaprefix")
@click.argument("new_metaprefix")
def main(old_metaprefix: str, new_metaprefix: str):
    """Rename a metaprefix."""
    registry = json.loads(BIOREGISTRY_PATH.read_text())
    for value in registry.values():
        if old_metaprefix in value:
            value[new_metaprefix] = value.pop(old_metaprefix)
        mappings = value.get("mappings")
        if mappings and old_metaprefix in mappings:
            mappings[new_metaprefix] = mappings.pop(old_metaprefix)

    write_registry(
        {prefix: Resource(prefix=prefix, **values) for prefix, values in registry.items()}
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Utilities for interacting with data and the schema."""

import json
import logging
from collections import defaultdict
from functools import lru_cache
from pathlib import Path
from typing import Dict, Mapping, Set, Union

from .constants import (
    BIOREGISTRY_PATH,
    COLLECTIONS_PATH,
    CONTEXTS_PATH,
    METAREGISTRY_PATH,
    MISMATCH_PATH,
)
from .schema import Attributable, Collection, Context, Registry, Resource
from .utils import extended_encoder

logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def read_metaregistry() -> Mapping[str, Registry]:
    """Read the metaregistry."""
    with open(METAREGISTRY_PATH, encoding="utf-8") as file:
        data = json.load(file)
    return {
        registry.prefix: registry
        for registry in (Registry(**record) for record in data["metaregistry"])
    }


@lru_cache(maxsize=1)
def read_registry() -> Mapping[str, Resource]:
    """Read the Bioregistry as JSON."""
    return _registry_from_path(BIOREGISTRY_PATH)


def _registry_from_path(path: Union[str, Path]) -> Mapping[str, Resource]:
    with open(path, encoding="utf-8") as file:
        data = json.load(file)
    return {prefix: Resource(prefix=prefix, **value) for prefix, value in data.items()}


def add_resource(resource: Resource) -> None:
    """Add a resource to the registry.

    :param resource: A resource object to write
    :raises KeyError: if the prefix is already present in the registry
    """
    registry = dict(read_registry())
    if resource.prefix in registry:
        raise KeyError("Tried to add duplicate entry to the registry")
    registry[resource.prefix] = resource
    # Clear the cache
    read_registry.cache_clear()
    write_registry(registry)


@lru_cache(maxsize=1)
def read_mismatches() -> Mapping[str, Mapping[str, str]]:
    """Read the mismatches as JSON."""
    with MISMATCH_PATH.open() as file:
        return json.load(file)


def is_mismatch(bioregistry_prefix, external_metaprefix, external_prefix) -> bool:
    """Return if the triple is a mismatch."""
    return external_prefix in read_mismatches().get(bioregistry_prefix, {}).get(
        external_metaprefix, {}
    )


@lru_cache(maxsize=1)
def read_collections() -> Mapping[str, Collection]:
    """Read the manually curated collections."""
    with open(COLLECTIONS_PATH, encoding="utf-8") as file:
        data = json.load(file)
    return {
        collection.identifier: collection
        for collection in (Collection(**record) for record in data["collections"])
    }


def write_collections(collections: Mapping[str, Collection]) -> None:
    """Write the collections."""
    values = [v for _, v in sorted(collections.items())]
    for collection in values:
        collection.resources = sorted(set(collection.resources))
    with open(COLLECTIONS_PATH, encoding="utf-8", mode="w") as file:
        json.dump(
            {"collections": values},
            file,
            indent=2,
            sort_keys=True,
            ensure_ascii=False,
            default=extended_encoder,
        )


def write_registry(registry: Mapping[str, Resource]):
    """Write to the Bioregistry."""
    with open(BIOREGISTRY_PATH, mode="w", encoding="utf-8") as file:
        json.dump(
            registry, file, indent=2, sort_keys=True, ensure_ascii=False, default=extended_encoder
        )


def write_metaregistry(metaregistry: Mapping[str, Registry]) -> None:
    """Write to the metaregistry."""
    values = [v for _, v in sorted(metaregistry.items())]
    with open(METAREGISTRY_PATH, mode="w", encoding="utf-8") as file:
        json.dump(
            {"metaregistry": values},
            fp=file,
            indent=2,
            sort_keys=True,
            ensure_ascii=False,
            default=extended_encoder,
        )


def write_contexts(contexts: Mapping[str, Context]) -> None:
    """Write to contexts."""
    with open(CONTEXTS_PATH, mode="w", encoding="utf-8") as file:
        json.dump(
            contexts,
            fp=file,
            indent=2,
            sort_keys=True,
            ensure_ascii=False,
            default=extended_encoder,
        )


def read_contributors() -> Mapping[str, Attributable]:
    """Get a mapping from contributor ORCID identifiers to author objects."""
    rv: Dict[str, Attributable] = {}
    for resource in read_registry().values():
        if resource.contributor and resource.contributor.orcid:
            rv[resource.contributor.orcid] = resource.contributor
        for contributor in resource.contributor_extras or []:
            if contributor.orcid:
                rv[contributor.orcid] = contributor
        if resource.reviewer and resource.reviewer.orcid:
            rv[resource.reviewer.orcid] = resource.reviewer
        contact = resource.get_contact()
        if contact and contact.orcid:
            rv[contact.orcid] = contact
    for metaresource in read_metaregistry().values():
        if metaresource.contact.orcid:
            rv[metaresource.contact.orcid] = metaresource.contact
    for collection in read_collections().values():
        for author in collection.authors or []:
            if author.orcid:
                rv[author.orcid] = author
    return rv


def read_prefix_contributions() -> Mapping[str, Set[str]]:
    """Get a mapping from contributor ORCID identifiers to prefixes."""
    rv = defaultdict(set)
    for prefix, resource in read_registry().items():
        if resource.contributor and resource.contributor.orcid:
            rv[resource.contributor.orcid].add(prefix)
        for contributor in resource.contributor_extras or []:
            if contributor.orcid:
                rv[contributor.orcid].add(prefix)
    return dict(rv)


def read_prefix_reviews() -> Mapping[str, Set[str]]:
    """Get a mapping from reviewer ORCID identifiers to prefixes."""
    rv = defaultdict(set)
    for prefix, resource in read_registry().items():
        if resource.reviewer and resource.reviewer.orcid:
            rv[resource.reviewer.orcid].add(prefix)
    return dict(rv)


def read_prefix_contacts() -> Mapping[str, Set[str]]:
    """Get a mapping from contact ORCID identifiers to prefixes."""
    rv = defaultdict(set)
    for prefix, resource in read_registry().items():
        contact_orcid = resource.get_contact_orcid()
        if contact_orcid:
            rv[contact_orcid].add(prefix)
    return dict(rv)


def read_collections_contributions() -> Mapping[str, Set[str]]:
    """Get a mapping from contributor ORCID identifiers to collections."""
    rv = defaultdict(set)
    for collection_id, resource in read_collections().items():
        for author in resource.authors or []:
            rv[author.orcid].add(collection_id)
    return dict(rv)


def read_registry_contributions() -> Mapping[str, Set[str]]:
    """Get a mapping from contributor ORCID identifiers to collections."""
    rv = defaultdict(set)
    for metaprefix, resource in read_metaregistry().items():
        if resource.contact and resource.contact.orcid:
            rv[resource.contact.orcid].add(metaprefix)
    return dict(rv)


def read_context_contributions() -> Mapping[str, Set[str]]:
    """Get a mapping from contributor ORCID identifiers to contexts."""
    rv = defaultdict(set)
    for context_key, context in read_contexts().items():
        for maintainer in context.maintainers:
            rv[maintainer.orcid].add(context_key)
    return dict(rv)


def read_contexts() -> Mapping[str, Context]:
    """Get a mapping from context keys to contexts."""
    return {
        key: Context(**data)
        for key, data in json.loads(CONTEXTS_PATH.read_text(encoding="utf-8")).items()
    }



"""Remove uninformative suffixes in names of resources."""

import click
from tabulate import tabulate

import bioregistry

suffixes = ["id", "accession"]


@click.command()
def _main():
    rows = []
    registry = bioregistry.read_registry()
    for prefix, resource in registry.items():
        name = bioregistry.get_name(prefix)
        for suffix in suffixes:
            if name.lower().endswith(f" {suffix}"):
                resource.name = name[: -len(suffix) - 1]
                rows.append((prefix, name))
    click.echo(tabulate(rows, headers=["prefix", "name"], tablefmt="github"))
    bioregistry.write_registry(registry)


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Align Prefix Commons with the Bioregistry."""

from typing import Sequence

from bioregistry.align.utils import Aligner
from bioregistry.external.prefixcommons import get_prefixcommons

__all__ = [
    "PrefixCommonsAligner",
]


class PrefixCommonsAligner(Aligner):
    """Aligner for Prefix Commons."""

    key = "prefixcommons"
    getter = get_prefixcommons
    curation_header = (
        "name",
        "miriam",
        "bioportal",
        "description",
        "example",
        "pattern",
        "uri_format",
    )
    # TODO consider updating
    include_new = False

    def get_curation_row(self, external_id, external_entry) -> Sequence[str]:
        """Prepare curation rows for unaligned Prefix Commons registry entries."""
        return [
            external_entry["name"],
            external_entry.get("miriam", ""),
            external_entry.get("bioportal", ""),
            external_entry.get("description", ""),
            external_entry.get("example", ""),
            external_entry.get("pattern", ""),
            external_entry.get("uri_format", ""),
        ]


if __name__ == "__main__":
    PrefixCommonsAligner.align()



# -*- coding: utf-8 -*-

"""Download BioContext."""

import json
from typing import Any, Mapping

from pystow.utils import download

from bioregistry.constants import EXTERNAL, URI_FORMAT_KEY

__all__ = [
    "get_biocontext",
]

DIRECTORY = EXTERNAL / "biocontext"
DIRECTORY.mkdir(exist_ok=True, parents=True)
RAW_PATH = DIRECTORY / "raw.json"
PROCESSED_PATH = DIRECTORY / "processed.json"
URL = "https://raw.githubusercontent.com/prefixcommons/biocontext/master/registry/commons_context.jsonld"


def get_biocontext(force_download: bool = False) -> Mapping[str, Mapping[str, Any]]:
    """Get the BioContext context map.

    :param force_download: If true, forces download. If false and the file
        is already cached, resuses it.
    :returns: The biocontext data dictionary

    .. seealso:: https://github.com/prefixcommons/biocontext
    """
    if PROCESSED_PATH.exists() and not force_download:
        with PROCESSED_PATH.open() as file:
            return json.load(file)
    download(url=URL, path=RAW_PATH, force=True)
    with RAW_PATH.open() as file:
        data = json.load(file)
    rv = {
        prefix: {URI_FORMAT_KEY: f"{uri_prefix.strip()}$1"}
        for prefix, uri_prefix in data["@context"].items()
    }
    with PROCESSED_PATH.open("w") as file:
        json.dump(rv, file, indent=2, sort_keys=True)
    return rv


if __name__ == "__main__":
    print(len(get_biocontext(force_download=True)))  # noqa:T201



# -*- coding: utf-8 -*-

"""API for collections."""

from typing import Optional

from .schema import Collection, Context
from .schema_utils import read_collections, read_contexts

__all__ = [
    "get_collection",
    "get_context",
]


def get_collection(identifier: str) -> Optional[Collection]:
    """Get the collection entry for the given identifier."""
    return read_collections().get(identifier)


def get_context(identifier: str) -> Optional[Context]:
    """Get the context for the given identifier."""
    return read_contexts().get(identifier)
